#   Copyright 2012 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Quota action implementations"""

import argparse
import itertools
import logging
import sys

from osc_lib.command import command
from osc_lib import utils

from openstackclient.i18n import _
from openstackclient.network import common

LOG = logging.getLogger(__name__)

# List the quota items, map the internal argument name to the option
# name that the user sees.

COMPUTE_QUOTAS = {
    'cores': 'cores',
    'fixed_ips': 'fixed-ips',
    'injected_file_content_bytes': 'injected-file-size',
    'injected_file_path_bytes': 'injected-path-size',
    'injected_files': 'injected-files',
    'instances': 'instances',
    'key_pairs': 'key-pairs',
    'metadata_items': 'properties',
    'ram': 'ram',
    'server_groups': 'server-groups',
    'server_group_members': 'server-group-members',
}

VOLUME_QUOTAS = {
    'backups': 'backups',
    'backup_gigabytes': 'backup-gigabytes',
    'gigabytes': 'gigabytes',
    'per_volume_gigabytes': 'per-volume-gigabytes',
    'snapshots': 'snapshots',
    'volumes': 'volumes',
}

IMPACT_VOLUME_TYPE_QUOTAS = [
    'gigabytes',
    'snapshots',
    'volumes',
]

NOVA_NETWORK_QUOTAS = {
    'floating_ips': 'floating-ips',
    'security_group_rules': 'secgroup-rules',
    'security_groups': 'secgroups',
}

NETWORK_QUOTAS = {
    'floatingip': 'floating-ips',
    'security_group_rule': 'secgroup-rules',
    'security_group': 'secgroups',
    'network': 'networks',
    'subnet': 'subnets',
    'port': 'ports',
    'router': 'routers',
    'rbac_policy': 'rbac-policies',
    'subnetpool': 'subnetpools',
}

NETWORK_KEYS = [
    'floating_ips',
    'networks',
    'rbac_policies',
    'routers',
    'ports',
    'security_group_rules',
    'security_groups',
    'subnet_pools',
    'subnets',
]


def _xform_get_quota(data, value, keys):
    res = []
    res_info = {}
    for key in keys:
        res_info[key] = getattr(data, key, '')

    res_info['id'] = value
    res.append(res_info)
    return res


def get_project(app, project):
    if project is not None:
        identity_client = app.client_manager.identity
        project = utils.find_resource(
            identity_client.projects,
            project,
        )
        project_id = project.id
        project_name = project.name
    elif app.client_manager.auth_ref:
        # Get the project from the current auth
        project = app.client_manager.auth_ref
        project_id = project.project_id
        project_name = project.project_name
    else:
        project_id = None
        project_name = None

    return {
        'id': project_id,
        'name': project_name,
    }


def get_compute_quotas(
    app,
    project_id,
    *,
    quota_class=False,
    detail=False,
    default=False,
):
    try:
        client = app.client_manager.compute
        if quota_class:
            # NOTE(stephenfin): The 'project' argument here could be anything
            # as the nova API doesn't care what you pass in. We only pass the
            # project in to avoid weirding people out :)
            quota = client.quota_classes.get(project_id)
        elif default:
            quota = client.quotas.defaults(project_id)
        else:
            quota = client.quotas.get(project_id, detail=detail)
    except Exception as e:
        if type(e).__name__ == 'EndpointNotFound':
            return {}
        raise
    return quota._info


def get_volume_quotas(
    app,
    project_id,
    *,
    quota_class=False,
    detail=False,
    default=False,
):
    try:
        client = app.client_manager.volume
        if quota_class:
            quota = client.quota_classes.get(project_id)
        elif default:
            quota = client.quotas.defaults(project_id)
        else:
            quota = client.quotas.get(project_id, usage=detail)
    except Exception as e:
        if type(e).__name__ == 'EndpointNotFound':
            return {}
        else:
            raise
    return quota._info


def get_network_quotas(
    app,
    project_id,
    *,
    quota_class=False,
    detail=False,
    default=False,
):
    def _network_quota_to_dict(network_quota, detail=False):
        if type(network_quota) is not dict:
            dict_quota = network_quota.to_dict()
        else:
            dict_quota = network_quota

        result = {}

        for key, values in dict_quota.items():
            if values is None:
                continue

            # NOTE(slaweq): Neutron returns values with key "used" but Nova for
            # example returns same data with key "in_use" instead. Because of
            # that we need to convert Neutron key to the same as is returned
            # from Nova to make result more consistent
            if isinstance(values, dict) and 'used' in values:
                values['in_use'] = values.pop("used")

            result[key] = values

        return result

    # neutron doesn't have the concept of quota classes and if we're using
    # nova-network we already fetched this
    if quota_class:
        return {}

    # we have nothing to return if we are not using neutron
    if not app.client_manager.is_network_endpoint_enabled():
        return {}

    client = app.client_manager.network
    if default:
        network_quota = client.get_quota_default(project_id)
        network_quota = _network_quota_to_dict(network_quota)
    else:
        network_quota = client.get_quota(project_id, details=detail)
        network_quota = _network_quota_to_dict(network_quota, detail=detail)
    return network_quota


class ListQuota(command.Lister):
    _description = _(
        "List quotas for all projects with non-default quota values or "
        "list detailed quota information for requested project"
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        # TODO(stephenfin): Remove in OSC 8.0
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                "**Deprecated** List quotas for this project <project> "
                "(name or ID). "
                "Use 'quota show' instead."
            ),
        )
        # TODO(stephenfin): Remove in OSC 8.0
        parser.add_argument(
            '--detail',
            dest='detail',
            action='store_true',
            default=False,
            help=_(
                "**Deprecated** Show details about quotas usage. "
                "Use 'quota show --usage' instead."
            ),
        )
        option = parser.add_mutually_exclusive_group(required=True)
        option.add_argument(
            '--compute',
            action='store_true',
            default=False,
            help=_('List compute quota'),
        )
        option.add_argument(
            '--volume',
            action='store_true',
            default=False,
            help=_('List volume quota'),
        )
        option.add_argument(
            '--network',
            action='store_true',
            default=False,
            help=_('List network quota'),
        )
        return parser

    def _get_detailed_quotas(self, parsed_args):
        project_info = get_project(self.app, parsed_args.project)
        project = project_info['id']

        quotas = {}

        if parsed_args.compute:
            quotas.update(
                get_compute_quotas(
                    self.app,
                    project,
                    detail=parsed_args.detail,
                )
            )

        if parsed_args.network:
            quotas.update(
                get_network_quotas(
                    self.app,
                    project,
                    detail=parsed_args.detail,
                )
            )

        if parsed_args.volume:
            quotas.update(
                get_volume_quotas(
                    self.app,
                    parsed_args,
                    detail=parsed_args.detail,
                ),
            )

        result = []
        for resource, values in quotas.items():
            # NOTE(slaweq): there is no detailed quotas info for some resources
            # and it shouldn't be displayed here
            if isinstance(values, dict):
                result.append(
                    {
                        'resource': resource,
                        'in_use': values.get('in_use'),
                        'reserved': values.get('reserved'),
                        'limit': values.get('limit'),
                    }
                )

        columns = (
            'resource',
            'in_use',
            'reserved',
            'limit',
        )
        column_headers = (
            'Resource',
            'In Use',
            'Reserved',
            'Limit',
        )

        return (
            column_headers,
            (utils.get_dict_properties(s, columns) for s in result),
        )

    def take_action(self, parsed_args):
        if parsed_args.detail:
            msg = _(
                "The --detail option has been deprecated. "
                "Use 'openstack quota show --usage' instead."
            )
            self.log.warning(msg)
        elif parsed_args.project:  # elif to avoid being too noisy
            msg = _(
                "The --project option has been deprecated. "
                "Use 'openstack quota show' instead."
            )
            self.log.warning(msg)

        result = []
        project_ids = []
        if parsed_args.project is None:
            for p in self.app.client_manager.identity.projects.list():
                project_ids.append(getattr(p, 'id', ''))
        else:
            identity_client = self.app.client_manager.identity
            project = utils.find_resource(
                identity_client.projects,
                parsed_args.project,
            )
            project_ids.append(getattr(project, 'id', ''))

        if parsed_args.compute:
            if parsed_args.detail:
                return self._get_detailed_quotas(parsed_args)

            compute_client = self.app.client_manager.compute
            for p in project_ids:
                try:
                    data = compute_client.quotas.get(p)
                except Exception as ex:
                    if (
                        type(ex).__name__ == 'NotFound'
                        or ex.http_status >= 400
                        and ex.http_status <= 499
                    ):
                        # Project not found, move on to next one
                        LOG.warning("Project %s not found: %s" % (p, ex))
                        continue
                    else:
                        raise

                result_data = _xform_get_quota(
                    data,
                    p,
                    COMPUTE_QUOTAS.keys(),
                )
                default_data = compute_client.quotas.defaults(p)
                result_default = _xform_get_quota(
                    default_data,
                    p,
                    COMPUTE_QUOTAS.keys(),
                )
                if result_default != result_data:
                    result += result_data

            columns = (
                'id',
                'cores',
                'fixed_ips',
                'injected_files',
                'injected_file_content_bytes',
                'injected_file_path_bytes',
                'instances',
                'key_pairs',
                'metadata_items',
                'ram',
                'server_groups',
                'server_group_members',
            )
            column_headers = (
                'Project ID',
                'Cores',
                'Fixed IPs',
                'Injected Files',
                'Injected File Content Bytes',
                'Injected File Path Bytes',
                'Instances',
                'Key Pairs',
                'Metadata Items',
                'Ram',
                'Server Groups',
                'Server Group Members',
            )
            return (
                column_headers,
                (utils.get_dict_properties(s, columns) for s in result),
            )

        if parsed_args.volume:
            if parsed_args.detail:
                return self._get_detailed_quotas(parsed_args)

            volume_client = self.app.client_manager.volume
            for p in project_ids:
                try:
                    data = volume_client.quotas.get(p)
                except Exception as ex:
                    if type(ex).__name__ == 'NotFound':
                        # Project not found, move on to next one
                        LOG.warning("Project %s not found: %s" % (p, ex))
                        continue
                    else:
                        raise

                result_data = _xform_get_quota(
                    data,
                    p,
                    VOLUME_QUOTAS.keys(),
                )
                default_data = volume_client.quotas.defaults(p)
                result_default = _xform_get_quota(
                    default_data,
                    p,
                    VOLUME_QUOTAS.keys(),
                )
                if result_default != result_data:
                    result += result_data

            columns = (
                'id',
                'backups',
                'backup_gigabytes',
                'gigabytes',
                'per_volume_gigabytes',
                'snapshots',
                'volumes',
            )
            column_headers = (
                'Project ID',
                'Backups',
                'Backup Gigabytes',
                'Gigabytes',
                'Per Volume Gigabytes',
                'Snapshots',
                'Volumes',
            )

            return (
                column_headers,
                (utils.get_dict_properties(s, columns) for s in result),
            )

        if parsed_args.network:
            if parsed_args.detail:
                return self._get_detailed_quotas(parsed_args)

            client = self.app.client_manager.network
            for p in project_ids:
                try:
                    data = client.get_quota(p)
                except Exception as ex:
                    if type(ex).__name__ == 'NotFound':
                        # Project not found, move on to next one
                        LOG.warning("Project %s not found: %s" % (p, ex))
                        continue
                    else:
                        raise

                result_data = _xform_get_quota(
                    data,
                    p,
                    NETWORK_KEYS,
                )
                default_data = client.get_quota_default(p)
                result_default = _xform_get_quota(
                    default_data,
                    p,
                    NETWORK_KEYS,
                )
                if result_default != result_data:
                    result += result_data

            columns = (
                'id',
                'floating_ips',
                'networks',
                'ports',
                'rbac_policies',
                'routers',
                'security_groups',
                'security_group_rules',
                'subnets',
                'subnet_pools',
            )
            column_headers = (
                'Project ID',
                'Floating IPs',
                'Networks',
                'Ports',
                'RBAC Policies',
                'Routers',
                'Security Groups',
                'Security Group Rules',
                'Subnets',
                'Subnet Pools',
            )

            return (
                column_headers,
                (utils.get_dict_properties(s, columns) for s in result),
            )

        return ((), ())


class SetQuota(common.NetDetectionMixin, command.Command):
    _description = _("Set quotas for project or class")

    def _build_options_list(self):
        help_fmt = _('New value for the %s quota')
        # Compute and volume quota options are always the same
        rets = [
            (k, v, help_fmt % v)
            for k, v in itertools.chain(
                COMPUTE_QUOTAS.items(),
                VOLUME_QUOTAS.items(),
            )
        ]
        # For docs build, we want to produce helps for both neutron and
        # nova-network options. They overlap, so we have to figure out which
        # need to be tagged as specific to one network type or the other.
        if self.is_docs_build:
            # NOTE(efried): This takes advantage of the fact that we know the
            # nova-net options are a subset of the neutron options. If that
            # ever changes, this algorithm will need to be adjusted accordingly
            inv_compute = set(NOVA_NETWORK_QUOTAS.values())
            for k, v in NETWORK_QUOTAS.items():
                _help = help_fmt % v
                if v not in inv_compute:
                    # This one is unique to neutron
                    _help = self.enhance_help_neutron(_help)
                rets.append((k, v, _help))
        elif self.is_neutron:
            rets.extend(
                [(k, v, help_fmt % v) for k, v in NETWORK_QUOTAS.items()]
            )
        elif self.is_nova_network:
            rets.extend(
                [(k, v, help_fmt % v) for k, v in NOVA_NETWORK_QUOTAS.items()]
            )
        return rets

    def get_parser(self, prog_name):
        parser = super(SetQuota, self).get_parser(prog_name)
        parser.add_argument(
            'project',
            metavar='<project/class>',
            help=_('Set quotas for this project or class (name or ID)'),
        )
        # TODO(stephenfin): Remove in OSC 8.0
        parser.add_argument(
            '--class',
            dest='quota_class',
            action='store_true',
            default=False,
            help=_(
                '**Deprecated** Set quotas for <class>. '
                'Deprecated as quota classes were never fully implemented '
                'and only the default class is supported. '
                '(compute and volume only)'
            ),
        )
        for k, v, h in self._build_options_list():
            parser.add_argument(
                '--%s' % v,
                metavar='<%s>' % v,
                dest=k,
                type=int,
                help=h,
            )
        parser.add_argument(
            '--volume-type',
            metavar='<volume-type>',
            help=_('Set quotas for a specific <volume-type>'),
        )
        force_group = parser.add_mutually_exclusive_group()
        force_group.add_argument(
            '--force',
            action='store_true',
            dest='force',
            # TODO(stephenfin): Change the default to False in Z or later
            default=None,
            help=_(
                'Force quota update (only supported by compute and network) '
                '(default for network)'
            ),
        )
        force_group.add_argument(
            '--no-force',
            action='store_false',
            dest='force',
            default=None,
            help=_(
                'Do not force quota update '
                '(only supported by compute and network) '
                '(default for compute)'
            ),
        )
        # kept here for backwards compatibility/to keep the neutron folks happy
        force_group.add_argument(
            '--check-limit',
            action='store_false',
            dest='force',
            default=None,
            help=argparse.SUPPRESS,
        )
        return parser

    def take_action(self, parsed_args):
        if parsed_args.quota_class:
            msg = _(
                "The '--class' option has been deprecated. Quota classes were "
                "never fully implemented and the compute and volume services "
                "only support a single 'default' quota class while the "
                "network service does not support quota classes at all. "
                "Please use 'openstack quota show --default' instead."
            )
            self.log.warning(msg)

        identity_client = self.app.client_manager.identity
        compute_client = self.app.client_manager.compute
        volume_client = self.app.client_manager.volume
        compute_kwargs = {}
        for k, v in COMPUTE_QUOTAS.items():
            value = getattr(parsed_args, k, None)
            if value is not None:
                compute_kwargs[k] = value

        if parsed_args.force is not None:
            compute_kwargs['force'] = parsed_args.force

        volume_kwargs = {}
        for k, v in VOLUME_QUOTAS.items():
            value = getattr(parsed_args, k, None)
            if value is not None:
                if parsed_args.volume_type and k in IMPACT_VOLUME_TYPE_QUOTAS:
                    k = k + '_%s' % parsed_args.volume_type
                volume_kwargs[k] = value

        network_kwargs = {}
        if parsed_args.force is True:
            # Unlike compute, network doesn't provide a simple boolean option.
            # Instead, it provides two options: 'force' and 'check_limit'
            # (a.k.a. 'not force')
            network_kwargs['force'] = True
        elif parsed_args.force is False:
            network_kwargs['check_limit'] = True
        else:
            msg = _(
                "This command currently defaults to '--force' when modifying "
                "network quotas. This behavior will change in a future "
                "release. Consider explicitly providing '--force' or "
                "'--no-force' options to avoid changes in behavior."
            )
            self.log.warning(msg)

        if self.app.client_manager.is_network_endpoint_enabled():
            for k, v in NETWORK_QUOTAS.items():
                value = getattr(parsed_args, k, None)
                if value is not None:
                    network_kwargs[k] = value
        else:
            for k, v in NOVA_NETWORK_QUOTAS.items():
                value = getattr(parsed_args, k, None)
                if value is not None:
                    compute_kwargs[k] = value

        if parsed_args.quota_class:
            if compute_kwargs:
                compute_client.quota_classes.update(
                    parsed_args.project,
                    **compute_kwargs,
                )
            if volume_kwargs:
                volume_client.quota_classes.update(
                    parsed_args.project,
                    **volume_kwargs,
                )
            if network_kwargs:
                sys.stderr.write(
                    "Network quotas are ignored since quota classes are not "
                    "supported."
                )
        else:
            project = utils.find_resource(
                identity_client.projects,
                parsed_args.project,
            ).id

            if compute_kwargs:
                compute_client.quotas.update(project, **compute_kwargs)
            if volume_kwargs:
                volume_client.quotas.update(project, **volume_kwargs)
            if (
                network_kwargs
                and self.app.client_manager.is_network_endpoint_enabled()
            ):
                network_client = self.app.client_manager.network
                network_client.update_quota(project, **network_kwargs)


class ShowQuota(command.Lister):
    _description = _("""Show quotas for project or class.

Specify ``--os-compute-api-version 2.50`` or higher to see ``server-groups``
and ``server-group-members`` output for a given quota class.""")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'project',
            metavar='<project/class>',
            nargs='?',
            help=_(
                'Show quotas for this project or class (name or ID) '
                '(defaults to current project)'
            ),
        )
        type_group = parser.add_mutually_exclusive_group()
        # TODO(stephenfin): Remove in OSC 8.0
        type_group.add_argument(
            '--class',
            dest='quota_class',
            action='store_true',
            default=False,
            help=_(
                '**Deprecated** Show quotas for <class>. '
                'Deprecated as quota classes were never fully implemented '
                'and only the default class is supported. '
                'Use --default instead which is also supported by the network '
                'service. '
                '(compute and volume only)'
            ),
        )
        type_group.add_argument(
            '--default',
            dest='default',
            action='store_true',
            default=False,
            help=_('Show default quotas for <project>'),
        )
        type_group.add_argument(
            '--usage',
            dest='usage',
            action='store_true',
            default=False,
            help=_('Show details about quotas usage'),
        )
        service_group = parser.add_mutually_exclusive_group()
        service_group.add_argument(
            '--all',
            action='store_const',
            const='all',
            dest='service',
            default='all',
            help=_('Show quotas for all services'),
        )
        service_group.add_argument(
            '--compute',
            action='store_const',
            const='compute',
            dest='service',
            default='all',
            help=_('Show compute quota'),
        )
        service_group.add_argument(
            '--volume',
            action='store_const',
            const='volume',
            dest='service',
            default='all',
            help=_('Show volume quota'),
        )
        service_group.add_argument(
            '--network',
            action='store_const',
            const='network',
            dest='service',
            default='all',
            help=_('Show network quota'),
        )

        return parser

    def take_action(self, parsed_args):
        project = parsed_args.project

        if parsed_args.quota_class:
            msg = _(
                "The '--class' option has been deprecated. Quota classes were "
                "never fully implemented and the compute and volume services "
                "only support a single 'default' quota class while the "
                "network service does not support quota classes at all. "
                "Please use 'openstack quota show --default' instead."
            )
            self.log.warning(msg)
        else:
            project_info = get_project(self.app, parsed_args.project)
            project = project_info['id']

        compute_quota_info = {}
        volume_quota_info = {}
        network_quota_info = {}

        # NOTE(stephenfin): These quota API calls do not validate the project
        # or class arguments and return what appears to be the default quota
        # values if the project or class does not exist. This is expected
        # behavior. However, we have already checked for the presence of the
        # project above so it shouldn't be an issue.
        if parsed_args.service in {'all', 'compute'}:
            compute_quota_info = get_compute_quotas(
                self.app,
                project,
                detail=parsed_args.usage,
                quota_class=parsed_args.quota_class,
                default=parsed_args.default,
            )
        if parsed_args.service in {'all', 'volume'}:
            volume_quota_info = get_volume_quotas(
                self.app,
                project,
                detail=parsed_args.usage,
                quota_class=parsed_args.quota_class,
                default=parsed_args.default,
            )
        if parsed_args.service in {'all', 'network'}:
            network_quota_info = get_network_quotas(
                self.app,
                project,
                detail=parsed_args.usage,
                quota_class=parsed_args.quota_class,
                default=parsed_args.default,
            )

        info = {}
        info.update(compute_quota_info)
        info.update(volume_quota_info)
        info.update(network_quota_info)

        # Map the internal quota names to the external ones
        # COMPUTE_QUOTAS and NETWORK_QUOTAS share floating-ips,
        # secgroup-rules and secgroups as dict value, so when
        # neutron is enabled, quotas of these three resources
        # in nova will be replaced by neutron's.
        for k, v in itertools.chain(
            COMPUTE_QUOTAS.items(),
            NOVA_NETWORK_QUOTAS.items(),
            VOLUME_QUOTAS.items(),
            NETWORK_QUOTAS.items(),
        ):
            if not k == v and info.get(k) is not None:
                info[v] = info[k]
                info.pop(k)

        # Remove the 'id' field since it's not very useful
        if 'id' in info:
            del info['id']

        # Remove the 'location' field for resources from openstacksdk
        if 'location' in info:
            del info['location']

        if not parsed_args.usage:
            result = [{'resource': k, 'limit': v} for k, v in info.items()]
        else:
            result = [{'resource': k, **v} for k, v in info.items()]

        columns = (
            'resource',
            'limit',
        )
        column_headers = (
            'Resource',
            'Limit',
        )

        if parsed_args.usage:
            columns += (
                'in_use',
                'reserved',
            )
            column_headers += (
                'In Use',
                'Reserved',
            )

        return (
            column_headers,
            (utils.get_dict_properties(s, columns) for s in result),
        )


class DeleteQuota(command.Command):
    _description = _(
        "Delete configured quota for a project and revert to defaults."
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'project',
            metavar='<project>',
            help=_('Delete quotas for this project (name or ID)'),
        )
        option = parser.add_mutually_exclusive_group()
        option.add_argument(
            '--all',
            action='store_const',
            const='all',
            dest='service',
            default='all',
            help=_('Delete project quotas for all services (default)'),
        )
        option.add_argument(
            '--compute',
            action='store_const',
            const='compute',
            dest='service',
            default='all',
            help=_(
                'Delete compute quotas for the project '
                '(including network quotas when using nova-network)'
            ),
        )
        option.add_argument(
            '--volume',
            action='store_const',
            const='volume',
            dest='service',
            default='all',
            help=_('Delete volume quotas for the project'),
        )
        option.add_argument(
            '--network',
            action='store_const',
            const='network',
            dest='service',
            default='all',
            help=_('Delete network quotas for the project'),
        )
        return parser

    def take_action(self, parsed_args):
        identity_client = self.app.client_manager.identity
        project = utils.find_resource(
            identity_client.projects,
            parsed_args.project,
        )

        # compute quotas
        if parsed_args.service in {'all', 'compute'}:
            compute_client = self.app.client_manager.compute
            compute_client.quotas.delete(project.id)

        # volume quotas
        if parsed_args.service in {'all', 'volume'}:
            volume_client = self.app.client_manager.volume
            volume_client.quotas.delete(project.id)

        # network quotas (but only if we're not using nova-network, otherwise
        # we already deleted the quotas in the compute step)
        if (
            parsed_args.service in {'all', 'network'}
            and self.app.client_manager.is_network_endpoint_enabled()
        ):
            network_client = self.app.client_manager.network
            network_client.delete_quota(project.id)

        return None



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Base API Library"""

from keystoneauth1 import exceptions as ks_exceptions
from keystoneauth1 import session as ks_session
from osc_lib import exceptions
import simplejson as json

from openstackclient.i18n import _


class KeystoneSession(object):
    """Wrapper for the Keystone Session

    Restore some requests.session.Session compatibility;
    keystoneauth1.session.Session.request() has the method and url
    arguments swapped from the rest of the requests-using world.

    """

    def __init__(
        self,
        session=None,
        endpoint=None,
        **kwargs
    ):
        """Base object that contains some common API objects and methods

        :param Session session:
            The default session to be used for making the HTTP API calls.
        :param string endpoint:
            The URL from the Service Catalog to be used as the base for API
            requests on this API.
        """

        super(KeystoneSession, self).__init__()

        # a requests.Session-style interface
        self.session = session
        self.endpoint = endpoint

    def _request(self, method, url, session=None, **kwargs):
        """Perform call into session

        All API calls are funneled through this method to provide a common
        place to finalize the passed URL and other things.

        :param string method:
            The HTTP method name, i.e. ``GET``, ``PUT``, etc
        :param string url:
            The API-specific portion of the URL path
        :param Session session:
            HTTP client session
        :param kwargs:
            keyword arguments passed to requests.request().
        :return: the requests.Response object
        """

        if not session:
            session = self.session
        if not session:
            session = ks_session.Session()

        if self.endpoint:
            if url:
                url = '/'.join([self.endpoint.rstrip('/'), url.lstrip('/')])
            else:
                url = self.endpoint.rstrip('/')

        # Why is ksc session backwards???
        return session.request(url, method, **kwargs)


class BaseAPI(KeystoneSession):
    """Base API"""

    def __init__(
        self,
        session=None,
        service_type=None,
        endpoint=None,
        **kwargs
    ):
        """Base object that contains some common API objects and methods

        :param Session session:
            The default session to be used for making the HTTP API calls.
        :param string service_type:
            API name, i.e. ``identity`` or ``compute``
        :param string endpoint:
            The URL from the Service Catalog to be used as the base for API
            requests on this API.
        """

        super(BaseAPI, self).__init__(session=session, endpoint=endpoint)

        self.service_type = service_type

    # The basic action methods all take a Session and return dict/lists

    def create(
        self,
        url,
        session=None,
        method=None,
        **params
    ):
        """Create a new resource

        :param string url:
            The API-specific portion of the URL path
        :param Session session:
            HTTP client session
        :param string method:
            HTTP method (default POST)
        """

        if not method:
            method = 'POST'
        ret = self._request(method, url, session=session, **params)
        # Should this move into _requests()?
        try:
            return ret.json()
        except json.JSONDecodeError:
            return ret

    def delete(
        self,
        url,
        session=None,
        **params
    ):
        """Delete a resource

        :param string url:
            The API-specific portion of the URL path
        :param Session session:
            HTTP client session
        """

        return self._request('DELETE', url, **params)

    def list(
        self,
        path,
        session=None,
        body=None,
        detailed=False,
        **params
    ):
        """Return a list of resources

        GET ${ENDPOINT}/${PATH}?${PARAMS}

        path is often the object's plural resource type

        :param string path:
            The API-specific portion of the URL path
        :param Session session:
            HTTP client session
        :param body: data that will be encoded as JSON and passed in POST
            request (GET will be sent by default)
        :param bool detailed:
            Adds '/details' to path for some APIs to return extended attributes
        :returns:
            JSON-decoded response, could be a list or a dict-wrapped-list
        """

        if detailed:
            path = '/'.join([path.rstrip('/'), 'details'])

        if body:
            ret = self._request(
                'POST',
                path,
                json=body,
                params=params,
            )
        else:
            ret = self._request(
                'GET',
                path,
                params=params,
            )
        try:
            return ret.json()
        except json.JSONDecodeError:
            return ret

    # Layered actions built on top of the basic action methods do not
    # explicitly take a Session but one may still be passed in kwargs

    def find_attr(
        self,
        path,
        value=None,
        attr=None,
        resource=None,
    ):
        """Find a resource via attribute or ID

        Most APIs return a list wrapped by a dict with the resource
        name as key.  Some APIs (Identity) return a dict when a query
        string is present and there is one return value.  Take steps to
        unwrap these bodies and return a single dict without any resource
        wrappers.

        :param string path:
            The API-specific portion of the URL path
        :param string value:
            value to search for
        :param string attr:
            attribute to use for resource search
        :param string resource:
            plural of the object resource name; defaults to path

        For example:
            n = find(netclient, 'network', 'networks', 'matrix')
        """

        # Default attr is 'name'
        if attr is None:
            attr = 'name'

        # Default resource is path - in many APIs they are the same
        if resource is None:
            resource = path

        def getlist(kw):
            """Do list call, unwrap resource dict if present"""
            ret = self.list(path, **kw)
            if isinstance(ret, dict) and resource in ret:
                ret = ret[resource]
            return ret

        # Search by attribute
        kwargs = {attr: value}
        data = getlist(kwargs)
        if isinstance(data, dict):
            return data
        if len(data) == 1:
            return data[0]
        if len(data) > 1:
            msg = _("Multiple %(resource)s exist with %(attr)s='%(value)s'")
            raise exceptions.CommandError(
                msg % {'resource': resource,
                       'attr': attr,
                       'value': value}
            )

        # Search by id
        kwargs = {'id': value}
        data = getlist(kwargs)
        if len(data) == 1:
            return data[0]
        msg = _("No %(resource)s with a %(attr)s or ID of '%(value)s' found")
        raise exceptions.CommandError(
            msg % {'resource': resource,
                   'attr': attr,
                   'value': value}
        )

    def find_bulk(
        self,
        path,
        **kwargs
    ):
        """Bulk load and filter locally

        :param string path:
            The API-specific portion of the URL path
        :param kwargs:
            A dict of AVPs to match - logical AND
        :returns: list of resource dicts
        """

        items = self.list(path)
        if isinstance(items, dict):
            # strip off the enclosing dict
            key = list(items.keys())[0]
            items = items[key]

        ret = []
        for o in items:
            try:
                if all(o[attr] == kwargs[attr] for attr in kwargs.keys()):
                    ret.append(o)
            except KeyError:
                continue

        return ret

    def find_one(
        self,
        path,
        **kwargs
    ):
        """Find a resource by name or ID

        :param string path:
            The API-specific portion of the URL path
        :returns:
            resource dict
        """

        bulk_list = self.find_bulk(path, **kwargs)
        num_bulk = len(bulk_list)
        if num_bulk == 0:
            msg = _("none found")
            raise exceptions.NotFound(msg)
        elif num_bulk > 1:
            msg = _("many found")
            raise RuntimeError(msg)
        return bulk_list[0]

    def find(
        self,
        path,
        value=None,
        attr=None,
    ):
        """Find a single resource by name or ID

        :param string path:
            The API-specific portion of the URL path
        :param string value:
            search expression
        :param string attr:
            name of attribute for secondary search
        """

        try:
            ret = self._request('GET', "/%s/%s" % (path, value)).json()
        except ks_exceptions.NotFound:
            kwargs = {attr: value}
            try:
                ret = self.find_one("/%s/detail" % (path), **kwargs)
            except ks_exceptions.NotFound:
                msg = _("%s not found") % value
                raise exceptions.NotFound(msg)

        return ret



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from openstackclient.common import limits
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes
from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes


class TestComputeLimits(compute_fakes.TestComputev2):

    absolute_columns = [
        'Name',
        'Value',
    ]

    rate_columns = [
        "Verb",
        "URI",
        "Value",
        "Remain",
        "Unit",
        "Next Available"
    ]

    def setUp(self):
        super().setUp()
        self.app.client_manager.volume_endpoint_enabled = False
        self.compute = self.app.client_manager.compute

        self.fake_limits = compute_fakes.FakeLimits()
        self.compute.limits.get.return_value = self.fake_limits

    def test_compute_show_absolute(self):
        arglist = ['--absolute']
        verifylist = [('is_absolute', True)]
        cmd = limits.ShowLimits(self.app, None)
        parsed_args = self.check_parser(cmd, arglist, verifylist)

        columns, data = cmd.take_action(parsed_args)

        ret_limits = list(data)
        compute_reference_limits = self.fake_limits.absolute_limits()

        self.assertEqual(self.absolute_columns, columns)
        self.assertEqual(compute_reference_limits, ret_limits)
        self.assertEqual(19, len(ret_limits))

    def test_compute_show_rate(self):
        arglist = ['--rate']
        verifylist = [('is_rate', True)]
        cmd = limits.ShowLimits(self.app, None)
        parsed_args = self.check_parser(cmd, arglist, verifylist)

        columns, data = cmd.take_action(parsed_args)

        ret_limits = list(data)
        compute_reference_limits = self.fake_limits.rate_limits()

        self.assertEqual(self.rate_columns, columns)
        self.assertEqual(compute_reference_limits, ret_limits)
        self.assertEqual(3, len(ret_limits))


class TestVolumeLimits(volume_fakes.TestVolume):
    absolute_columns = [
        'Name',
        'Value',
    ]

    rate_columns = [
        "Verb",
        "URI",
        "Value",
        "Remain",
        "Unit",
        "Next Available"
    ]

    def setUp(self):
        super().setUp()
        self.app.client_manager.compute_endpoint_enabled = False
        self.volume = self.app.client_manager.volume

        self.fake_limits = volume_fakes.FakeLimits()
        self.volume.limits.get.return_value = self.fake_limits

    def test_volume_show_absolute(self):
        arglist = ['--absolute']
        verifylist = [('is_absolute', True)]
        cmd = limits.ShowLimits(self.app, None)
        parsed_args = self.check_parser(cmd, arglist, verifylist)

        columns, data = cmd.take_action(parsed_args)

        ret_limits = list(data)
        compute_reference_limits = self.fake_limits.absolute_limits()

        self.assertEqual(self.absolute_columns, columns)
        self.assertEqual(compute_reference_limits, ret_limits)
        self.assertEqual(10, len(ret_limits))

    def test_volume_show_rate(self):
        arglist = ['--rate']
        verifylist = [('is_rate', True)]
        cmd = limits.ShowLimits(self.app, None)
        parsed_args = self.check_parser(cmd, arglist, verifylist)

        columns, data = cmd.take_action(parsed_args)

        ret_limits = list(data)
        compute_reference_limits = self.fake_limits.rate_limits()

        self.assertEqual(self.rate_columns, columns)
        self.assertEqual(compute_reference_limits, ret_limits)
        self.assertEqual(3, len(ret_limits))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from cinderclient import api_versions
from osc_lib import exceptions

from openstackclient.tests.unit import utils as tests_utils
from openstackclient.tests.unit.volume.v2 import fakes as v2_volume_fakes
from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import block_storage_manage


class TestBlockStorageManage(v2_volume_fakes.TestVolume):

    def setUp(self):
        super().setUp()

        self.volumes_mock = self.app.client_manager.volume.volumes
        self.volumes_mock.reset_mock()
        self.snapshots_mock = self.app.client_manager.volume.volume_snapshots
        self.snapshots_mock.reset_mock()


class TestBlockStorageVolumeManage(TestBlockStorageManage):

    volume_manage_list = volume_fakes.create_volume_manage_list_records()

    def setUp(self):
        super().setUp()

        self.volumes_mock.list_manageable.return_value = (
            self.volume_manage_list)

        # Get the command object to test
        self.cmd = block_storage_manage.BlockStorageManageVolumes(
            self.app, None)

    def test_block_storage_volume_manage_list(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.8')
        host = 'fake_host'
        arglist = [
            host,
        ]
        verifylist = [
            ('host', host),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'reference',
            'size',
            'safe_to_manage',
            'reason_not_safe',
            'cinder_id',
            'extra_info',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        datalist = []
        for volume_record in self.volume_manage_list:
            manage_details = (
                volume_record.reference,
                volume_record.size,
                volume_record.safe_to_manage,
                volume_record.reason_not_safe,
                volume_record.cinder_id,
                volume_record.extra_info,
            )
            datalist.append(manage_details)
        datalist = tuple(datalist)

        # confirming if all expected values are present in the result.
        self.assertEqual(datalist, tuple(data))

        # checking if proper call was made to get volume manageable list
        self.volumes_mock.list_manageable.assert_called_with(
            host=parsed_args.host,
            detailed=parsed_args.detailed,
            marker=parsed_args.marker,
            limit=parsed_args.limit,
            offset=parsed_args.offset,
            sort=parsed_args.sort,
            cluster=parsed_args.cluster,
        )

    def test_block_storage_volume_manage_pre_38(self):
        host = 'fake_host'
        arglist = [
            host,
        ]
        verifylist = [
            ('host', host),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(exceptions.CommandError, self.cmd.take_action,
                                parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.8 or greater is required', str(exc))

    def test_block_storage_volume_manage_pre_317(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.16')
        cluster = 'fake_cluster'
        arglist = [
            '--cluster', cluster,
        ]
        verifylist = [
            ('cluster', cluster),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(exceptions.CommandError, self.cmd.take_action,
                                parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.17 or greater is required', str(exc))
        self.assertIn('--cluster', str(exc))

    def test_block_storage_volume_manage_host_and_cluster(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.17')
        host = 'fake_host'
        cluster = 'fake_cluster'
        arglist = [
            host,
            '--cluster', cluster,
        ]
        verifylist = [
            ('host', host),
            ('cluster', cluster),
        ]
        exc = self.assertRaises(tests_utils.ParserException,
                                self.check_parser, self.cmd,
                                arglist, verifylist)
        self.assertIn(
            'argument --cluster: not allowed with argument <host>', str(exc))

    def test_block_storage_volume_manage_list_all_args(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.8')
        host = 'fake_host'
        detailed = True
        marker = 'fake_marker'
        limit = '5'
        offset = '3'
        sort = 'size:asc'
        arglist = [
            host,
            '--detailed', str(detailed),
            '--marker', marker,
            '--limit', limit,
            '--offset', offset,
            '--sort', sort,
        ]
        verifylist = [
            ('host', host),
            ('detailed', str(detailed)),
            ('marker', marker),
            ('limit', limit),
            ('offset', offset),
            ('sort', sort),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'reference',
            'size',
            'safe_to_manage',
            'reason_not_safe',
            'cinder_id',
            'extra_info',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        datalist = []
        for volume_record in self.volume_manage_list:
            manage_details = (
                volume_record.reference,
                volume_record.size,
                volume_record.safe_to_manage,
                volume_record.reason_not_safe,
                volume_record.cinder_id,
                volume_record.extra_info,
            )
            datalist.append(manage_details)
        datalist = tuple(datalist)

        # confirming if all expected values are present in the result.
        self.assertEqual(datalist, tuple(data))

        # checking if proper call was made to get volume manageable list
        self.volumes_mock.list_manageable.assert_called_with(
            host=host,
            detailed=detailed,
            marker=marker,
            limit=limit,
            offset=offset,
            sort=sort,
            cluster=parsed_args.cluster,
        )


class TestBlockStorageSnapshotManage(TestBlockStorageManage):

    snapshot_manage_list = volume_fakes.create_snapshot_manage_list_records()

    def setUp(self):
        super().setUp()

        self.snapshots_mock.list_manageable.return_value = (
            self.snapshot_manage_list)

        # Get the command object to test
        self.cmd = block_storage_manage.BlockStorageManageSnapshots(
            self.app, None)

    def test_block_storage_snapshot_manage_list(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.8')
        host = 'fake_host'
        arglist = [
            host,
        ]
        verifylist = [
            ('host', host),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'reference',
            'size',
            'safe_to_manage',
            'source_reference',
            'reason_not_safe',
            'cinder_id',
            'extra_info',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        datalist = []
        for snapshot_record in self.snapshot_manage_list:
            manage_details = (
                snapshot_record.reference,
                snapshot_record.size,
                snapshot_record.safe_to_manage,
                snapshot_record.source_reference,
                snapshot_record.reason_not_safe,
                snapshot_record.cinder_id,
                snapshot_record.extra_info,
            )
            datalist.append(manage_details)
        datalist = tuple(datalist)

        # confirming if all expected values are present in the result.
        self.assertEqual(datalist, tuple(data))

        # checking if proper call was made to get snapshot manageable list
        self.snapshots_mock.list_manageable.assert_called_with(
            host=parsed_args.host,
            detailed=parsed_args.detailed,
            marker=parsed_args.marker,
            limit=parsed_args.limit,
            offset=parsed_args.offset,
            sort=parsed_args.sort,
            cluster=parsed_args.cluster,
        )

    def test_block_storage_volume_manage_pre_38(self):
        host = 'fake_host'
        arglist = [
            host,
        ]
        verifylist = [
            ('host', host),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(exceptions.CommandError, self.cmd.take_action,
                                parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.8 or greater is required', str(exc))

    def test_block_storage_volume_manage_pre_317(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.16')
        cluster = 'fake_cluster'
        arglist = [
            '--cluster', cluster,
        ]
        verifylist = [
            ('cluster', cluster),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(exceptions.CommandError, self.cmd.take_action,
                                parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.17 or greater is required', str(exc))
        self.assertIn('--cluster', str(exc))

    def test_block_storage_volume_manage_host_and_cluster(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.17')
        host = 'fake_host'
        cluster = 'fake_cluster'
        arglist = [
            host,
            '--cluster', cluster,
        ]
        verifylist = [
            ('host', host),
            ('cluster', cluster),
        ]
        exc = self.assertRaises(tests_utils.ParserException,
                                self.check_parser, self.cmd,
                                arglist, verifylist)
        self.assertIn(
            'argument --cluster: not allowed with argument <host>', str(exc))

    def test_block_storage_volume_manage_list_all_args(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.8')
        host = 'fake_host'
        detailed = True
        marker = 'fake_marker'
        limit = '5'
        offset = '3'
        sort = 'size:asc'
        arglist = [
            host,
            '--detailed', str(detailed),
            '--marker', marker,
            '--limit', limit,
            '--offset', offset,
            '--sort', sort,
        ]
        verifylist = [
            ('host', host),
            ('detailed', str(detailed)),
            ('marker', marker),
            ('limit', limit),
            ('offset', offset),
            ('sort', sort),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'reference',
            'size',
            'safe_to_manage',
            'source_reference',
            'reason_not_safe',
            'cinder_id',
            'extra_info',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        datalist = []
        for snapshot_record in self.snapshot_manage_list:
            manage_details = (
                snapshot_record.reference,
                snapshot_record.size,
                snapshot_record.safe_to_manage,
                snapshot_record.source_reference,
                snapshot_record.reason_not_safe,
                snapshot_record.cinder_id,
                snapshot_record.extra_info,
            )
            datalist.append(manage_details)
        datalist = tuple(datalist)

        # confirming if all expected values are present in the result.
        self.assertEqual(datalist, tuple(data))

        # checking if proper call was made to get snapshot manageable list
        self.snapshots_mock.list_manageable.assert_called_with(
            host=host,
            detailed=detailed,
            marker=marker,
            limit=limit,
            offset=offset,
            sort=sort,
            cluster=parsed_args.cluster,
        )



# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from openstackclient.tests.functional.network.v2 import common


class NetworkQosRuleTestsMinimumBandwidth(common.NetworkTests):
    """Functional tests for QoS minimum bandwidth rule"""

    def setUp(self):
        super(NetworkQosRuleTestsMinimumBandwidth, self).setUp()
        # Nothing in this class works with Nova Network
        if not self.haz_network:
            self.skipTest("No Network service present")

        self.QOS_POLICY_NAME = 'qos_policy_%s' % uuid.uuid4().hex

        self.openstack(
            'network qos policy create %s' % self.QOS_POLICY_NAME
        )
        self.addCleanup(self.openstack,
                        'network qos policy delete %s' % self.QOS_POLICY_NAME)
        cmd_output = self.openstack(
            'network qos rule create '
            '--type minimum-bandwidth '
            '--min-kbps 2800 '
            '--egress %s' %
            self.QOS_POLICY_NAME,
            parse_output=True,
        )
        self.RULE_ID = cmd_output['id']
        self.addCleanup(self.openstack,
                        'network qos rule delete %s %s' %
                        (self.QOS_POLICY_NAME, self.RULE_ID))
        self.assertTrue(self.RULE_ID)

    def test_qos_rule_create_delete(self):
        # This is to check the output of qos rule delete
        policy_name = uuid.uuid4().hex
        self.openstack('network qos policy create %s' % policy_name)
        self.addCleanup(self.openstack,
                        'network qos policy delete %s' % policy_name)
        rule = self.openstack(
            'network qos rule create '
            '--type minimum-bandwidth '
            '--min-kbps 2800 '
            '--egress %s' % policy_name,
            parse_output=True,
        )
        raw_output = self.openstack(
            'network qos rule delete %s %s' %
            (policy_name, rule['id']))
        self.assertEqual('', raw_output)

    def test_qos_rule_list(self):
        cmd_output = self.openstack(
            'network qos rule list %s' % self.QOS_POLICY_NAME,
            parse_output=True,)
        self.assertIn(self.RULE_ID, [rule['ID'] for rule in cmd_output])

    def test_qos_rule_show(self):
        cmd_output = self.openstack(
            'network qos rule show %s %s' %
            (self.QOS_POLICY_NAME, self.RULE_ID),
            parse_output=True,
        )
        self.assertEqual(self.RULE_ID, cmd_output['id'])

    def test_qos_rule_set(self):
        self.openstack('network qos rule set --min-kbps 7500 %s %s' %
                       (self.QOS_POLICY_NAME, self.RULE_ID))
        cmd_output = self.openstack(
            'network qos rule show %s %s' %
            (self.QOS_POLICY_NAME, self.RULE_ID),
            parse_output=True,
        )
        self.assertEqual(7500, cmd_output['min_kbps'])


class NetworkQosRuleTestsMinimumPacketRate(common.NetworkTests):
    """Functional tests for QoS minimum packet rate rule"""

    def setUp(self):
        super(NetworkQosRuleTestsMinimumPacketRate, self).setUp()
        # Nothing in this class works with Nova Network
        if not self.haz_network:
            self.skipTest("No Network service present")

        self.QOS_POLICY_NAME = 'qos_policy_%s' % uuid.uuid4().hex

        self.openstack(
            'network qos policy create %s' % self.QOS_POLICY_NAME
        )
        self.addCleanup(self.openstack,
                        'network qos policy delete %s' % self.QOS_POLICY_NAME)
        cmd_output = self.openstack(
            'network qos rule create '
            '--type minimum-packet-rate '
            '--min-kpps 2800 '
            '--egress %s' %
            self.QOS_POLICY_NAME,
            parse_output=True,
        )
        self.RULE_ID = cmd_output['id']
        self.addCleanup(self.openstack,
                        'network qos rule delete %s %s' %
                        (self.QOS_POLICY_NAME, self.RULE_ID))
        self.assertTrue(self.RULE_ID)

    def test_qos_rule_create_delete(self):
        # This is to check the output of qos rule delete
        policy_name = uuid.uuid4().hex
        self.openstack('network qos policy create %s' % policy_name)
        self.addCleanup(self.openstack,
                        'network qos policy delete %s' % policy_name)
        rule = self.openstack(
            'network qos rule create '
            '--type minimum-packet-rate '
            '--min-kpps 2800 '
            '--egress %s' % policy_name,
            parse_output=True,
        )
        raw_output = self.openstack(
            'network qos rule delete %s %s' %
            (policy_name, rule['id']))
        self.assertEqual('', raw_output)

    def test_qos_rule_list(self):
        cmd_output = self.openstack(
            'network qos rule list %s' % self.QOS_POLICY_NAME,
            parse_output=True,)
        self.assertIn(self.RULE_ID, [rule['ID'] for rule in cmd_output])

    def test_qos_rule_show(self):
        cmd_output = self.openstack(
            'network qos rule show %s %s' %
            (self.QOS_POLICY_NAME, self.RULE_ID),
            parse_output=True,
        )
        self.assertEqual(self.RULE_ID, cmd_output['id'])

    def test_qos_rule_set(self):
        self.openstack('network qos rule set --min-kpps 7500 --any %s %s' %
                       (self.QOS_POLICY_NAME, self.RULE_ID))
        cmd_output = self.openstack(
            'network qos rule show %s %s' %
            (self.QOS_POLICY_NAME, self.RULE_ID),
            parse_output=True,
        )
        self.assertEqual(7500, cmd_output['min_kpps'])
        self.assertEqual('any', cmd_output['direction'])


class NetworkQosRuleTestsDSCPMarking(common.NetworkTests):
    """Functional tests for QoS DSCP marking rule"""

    def setUp(self):
        super(NetworkQosRuleTestsDSCPMarking, self).setUp()
        # Nothing in this class works with Nova Network
        if not self.haz_network:
            self.skipTest("No Network service present")

        self.QOS_POLICY_NAME = 'qos_policy_%s' % uuid.uuid4().hex
        self.openstack(
            'network qos policy create %s' % self.QOS_POLICY_NAME
        )
        self.addCleanup(self.openstack,
                        'network qos policy delete %s' % self.QOS_POLICY_NAME)
        cmd_output = self.openstack(
            'network qos rule create '
            '--type dscp-marking '
            '--dscp-mark 8 %s' %
            self.QOS_POLICY_NAME,
            parse_output=True,
        )
        self.RULE_ID = cmd_output['id']
        self.addCleanup(self.openstack,
                        'network qos rule delete %s %s' %
                        (self.QOS_POLICY_NAME, self.RULE_ID))
        self.assertTrue(self.RULE_ID)

    def test_qos_rule_create_delete(self):
        # This is to check the output of qos rule delete
        policy_name = uuid.uuid4().hex
        self.openstack('network qos policy create %s' % policy_name)
        self.addCleanup(self.openstack,
                        'network qos policy delete %s' % policy_name)
        rule = self.openstack(
            'network qos rule create '
            '--type dscp-marking '
            '--dscp-mark 8 %s' % policy_name,
            parse_output=True,
        )
        raw_output = self.openstack(
            'network qos rule delete %s %s' %
            (policy_name, rule['id']))
        self.assertEqual('', raw_output)

    def test_qos_rule_list(self):
        cmd_output = self.openstack(
            'network qos rule list %s' % self.QOS_POLICY_NAME,
            parse_output=True,)
        self.assertIn(self.RULE_ID, [rule['ID'] for rule in cmd_output])

    def test_qos_rule_show(self):
        cmd_output = self.openstack(
            'network qos rule show %s %s' %
            (self.QOS_POLICY_NAME, self.RULE_ID),
            parse_output=True,
        )
        self.assertEqual(self.RULE_ID, cmd_output['id'])

    def test_qos_rule_set(self):
        self.openstack('network qos rule set --dscp-mark 32 %s %s' %
                       (self.QOS_POLICY_NAME, self.RULE_ID))
        cmd_output = self.openstack(
            'network qos rule show %s %s' %
            (self.QOS_POLICY_NAME, self.RULE_ID),
            parse_output=True,
        )
        self.assertEqual(32, cmd_output['dscp_mark'])


class NetworkQosRuleTestsBandwidthLimit(common.NetworkTests):
    """Functional tests for QoS bandwidth limit rule"""

    def setUp(self):
        super(NetworkQosRuleTestsBandwidthLimit, self).setUp()
        # Nothing in this class works with Nova Network
        if not self.haz_network:
            self.skipTest("No Network service present")

        self.QOS_POLICY_NAME = 'qos_policy_%s' % uuid.uuid4().hex
        self.openstack(
            'network qos policy create %s' % self.QOS_POLICY_NAME
        )
        self.addCleanup(self.openstack,
                        'network qos policy delete %s' % self.QOS_POLICY_NAME)
        cmd_output = self.openstack(
            'network qos rule create '
            '--type bandwidth-limit '
            '--max-kbps 10000 '
            '--egress %s' %
            self.QOS_POLICY_NAME,
            parse_output=True,
        )
        self.RULE_ID = cmd_output['id']
        self.addCleanup(self.openstack,
                        'network qos rule delete %s %s' %
                        (self.QOS_POLICY_NAME, self.RULE_ID))
        self.assertTrue(self.RULE_ID)

    def test_qos_rule_create_delete(self):
        # This is to check the output of qos rule delete
        policy_name = uuid.uuid4().hex
        self.openstack('network qos policy create %s' % policy_name)
        self.addCleanup(self.openstack,
                        'network qos policy delete %s' % policy_name)
        rule = self.openstack(
            'network qos rule create '
            '--type bandwidth-limit '
            '--max-kbps 10000 '
            '--max-burst-kbits 1400 '
            '--egress %s' % policy_name,
            parse_output=True,
        )
        raw_output = self.openstack(
            'network qos rule delete %s %s' %
            (policy_name, rule['id']))
        self.assertEqual('', raw_output)

    def test_qos_rule_list(self):
        cmd_output = self.openstack(
            'network qos rule list %s' %
            self.QOS_POLICY_NAME,
            parse_output=True,)
        self.assertIn(self.RULE_ID, [rule['ID'] for rule in cmd_output])

    def test_qos_rule_show(self):
        cmd_output = self.openstack(
            'network qos rule show %s %s' %
            (self.QOS_POLICY_NAME, self.RULE_ID),
            parse_output=True,
        )
        self.assertEqual(self.RULE_ID, cmd_output['id'])

    def test_qos_rule_set(self):
        self.openstack('network qos rule set --max-kbps 15000 '
                       '--max-burst-kbits 1800 '
                       '--ingress %s %s' %
                       (self.QOS_POLICY_NAME, self.RULE_ID))
        cmd_output = self.openstack(
            'network qos rule show %s %s' %
            (self.QOS_POLICY_NAME, self.RULE_ID),
            parse_output=True,
        )
        self.assertEqual(15000, cmd_output['max_kbps'])
        self.assertEqual(1800, cmd_output['max_burst_kbps'])
        self.assertEqual('ingress', cmd_output['direction'])



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock

from openstackclient.identity.v3 import catalog
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit import utils


class TestCatalog(utils.TestCommand):

    fake_service = {
        'id': 'qwertyuiop',
        'type': 'compute',
        'name': 'supernova',
        'endpoints': [
            {
                'region': 'onlyone',
                'url': 'https://public.example.com',
                'interface': 'public',
            },
            {
                'region_id': 'onlyone',
                'url': 'https://admin.example.com',
                'interface': 'admin',
            },
            {
                'url': 'https://internal.example.com',
                'interface': 'internal',
            },
            {
                'region': None,
                'url': 'https://none.example.com',
                'interface': 'none',
            },
        ],
    }

    def setUp(self):
        super(TestCatalog, self).setUp()

        self.sc_mock = mock.Mock()
        self.sc_mock.service_catalog.catalog.return_value = [
            self.fake_service,
        ]

        self.auth_mock = mock.Mock()
        self.app.client_manager.session = self.auth_mock

        self.auth_mock.auth.get_auth_ref.return_value = self.sc_mock


class TestCatalogList(TestCatalog):

    def setUp(self):
        super(TestCatalogList, self).setUp()

        # Get the command object to test
        self.cmd = catalog.ListCatalog(self.app, None)

    def test_catalog_list(self):
        auth_ref = identity_fakes.fake_auth_ref(
            identity_fakes.TOKEN_WITH_PROJECT_ID,
            fake_service=self.fake_service,
        )
        self.ar_mock = mock.PropertyMock(return_value=auth_ref)
        type(self.app.client_manager).auth_ref = self.ar_mock

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        collist = ('Name', 'Type', 'Endpoints')
        self.assertEqual(collist, columns)
        datalist = ((
            'supernova',
            'compute',
            catalog.EndpointsColumn(
                auth_ref.service_catalog.catalog[0]['endpoints']),
        ), )
        self.assertCountEqual(datalist, tuple(data))


class TestCatalogShow(TestCatalog):

    def setUp(self):
        super(TestCatalogShow, self).setUp()

        # Get the command object to test
        self.cmd = catalog.ShowCatalog(self.app, None)

    def test_catalog_show(self):
        auth_ref = identity_fakes.fake_auth_ref(
            identity_fakes.TOKEN_WITH_PROJECT_ID,
            fake_service=self.fake_service,
        )
        self.ar_mock = mock.PropertyMock(return_value=auth_ref)
        type(self.app.client_manager).auth_ref = self.ar_mock

        arglist = [
            'compute',
        ]
        verifylist = [
            ('service', 'compute'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)

        collist = ('endpoints', 'id', 'name', 'type')
        self.assertEqual(collist, columns)
        datalist = (
            catalog.EndpointsColumn(
                auth_ref.service_catalog.catalog[0]['endpoints']),
            'qwertyuiop',
            'supernova',
            'compute',
        )
        self.assertCountEqual(datalist, data)


class TestFormatColumns(TestCatalog):
    def test_endpoints_column_human_readabale(self):
        col = catalog.EndpointsColumn(self.fake_service['endpoints'])
        self.assertEqual(
            'onlyone\n  public: https://public.example.com\n'
            'onlyone\n  admin: https://admin.example.com\n'
            '<none>\n  internal: https://internal.example.com\n'
            '<none>\n  none: https://none.example.com\n',
            col.human_readable())



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Volume v1 Volume action implementations"""

import argparse
import functools
import logging

from cliff import columns as cliff_columns
from osc_lib.cli import format_columns
from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class AttachmentsColumn(cliff_columns.FormattableColumn):
    """Formattable column for attachments column.

    Unlike the parent FormattableColumn class, the initializer of the
    class takes server_cache as the second argument.
    osc_lib.utils.get_item_properties instantiate cliff FormattableColumn
    object with a single parameter "column value", so you need to pass
    a partially initialized class like
    ``functools.partial(AttachmentsColumn, server_cache)``.
    """

    def __init__(self, value, server_cache=None):
        super(AttachmentsColumn, self).__init__(value)
        self._server_cache = server_cache or {}

    def human_readable(self):
        """Return a formatted string of a volume's attached instances

        :rtype: a string of formatted instances
        """

        msg = ''
        for attachment in self._value:
            server = attachment['server_id']
            if server in self._server_cache.keys():
                server = self._server_cache[server].name
            device = attachment['device']
            msg += 'Attached to %s on %s ' % (server, device)
        return msg


def _check_size_arg(args):
    """Check whether --size option is required or not.

    Require size parameter only in case when snapshot or source
    volume is not specified.
    """

    if ((args.snapshot or args.source)
            is None and args.size is None):
        msg = _("--size is a required option if snapshot "
                "or source volume is not specified.")
        raise exceptions.CommandError(msg)


class CreateVolume(command.ShowOne):
    _description = _("Create new volume")

    def get_parser(self, prog_name):
        parser = super(CreateVolume, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_('Volume name'),
        )
        parser.add_argument(
            '--size',
            metavar='<size>',
            type=int,
            help=_("Volume size in GB (Required unless --snapshot or "
                   "--source is specified)"),
        )
        parser.add_argument(
            '--type',
            metavar='<volume-type>',
            help=_("Set the type of volume"),
        )
        source_group = parser.add_mutually_exclusive_group()
        source_group.add_argument(
            '--image',
            metavar='<image>',
            help=_('Use <image> as source of volume (name or ID)'),
        )
        source_group.add_argument(
            '--snapshot',
            metavar='<snapshot>',
            help=_('Use <snapshot> as source of volume (name or ID)'),
        )
        source_group.add_argument(
            '--snapshot-id',
            metavar='<snapshot-id>',
            help=argparse.SUPPRESS,
        )
        source_group.add_argument(
            '--source',
            metavar='<volume>',
            help=_('Volume to clone (name or ID)'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Volume description'),
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_('Specify an alternate user (name or ID)'),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Specify an alternate project (name or ID)'),
        )
        parser.add_argument(
            '--availability-zone',
            metavar='<availability-zone>',
            help=_('Create volume in <availability-zone>'),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            help=_('Set a property on this volume '
                   '(repeat option to set multiple properties)'),
        )
        bootable_group = parser.add_mutually_exclusive_group()
        bootable_group.add_argument(
            "--bootable",
            action="store_true",
            help=_("Mark volume as bootable")
        )
        bootable_group.add_argument(
            "--non-bootable",
            action="store_true",
            help=_("Mark volume as non-bootable (default)")
        )
        readonly_group = parser.add_mutually_exclusive_group()
        readonly_group.add_argument(
            "--read-only",
            action="store_true",
            help=_("Set volume to read-only access mode")
        )
        readonly_group.add_argument(
            "--read-write",
            action="store_true",
            help=_("Set volume to read-write access mode (default)")
        )

        return parser

    def take_action(self, parsed_args):
        _check_size_arg(parsed_args)
        identity_client = self.app.client_manager.identity
        image_client = self.app.client_manager.image
        volume_client = self.app.client_manager.volume

        source_volume = None
        if parsed_args.source:
            source_volume = utils.find_resource(
                volume_client.volumes,
                parsed_args.source,
            ).id

        project = None
        if parsed_args.project:
            project = utils.find_resource(
                identity_client.tenants,
                parsed_args.project,
            ).id

        user = None
        if parsed_args.user:
            user = utils.find_resource(
                identity_client.users,
                parsed_args.user,
            ).id

        image = None
        if parsed_args.image:
            image = utils.find_resource(
                image_client.images,
                parsed_args.image,
            ).id

        snapshot = parsed_args.snapshot or parsed_args.snapshot_id

        volume = volume_client.volumes.create(
            parsed_args.size,
            snapshot,
            source_volume,
            parsed_args.name,
            parsed_args.description,
            parsed_args.type,
            user,
            project,
            parsed_args.availability_zone,
            parsed_args.property,
            image,
        )

        if parsed_args.bootable or parsed_args.non_bootable:
            try:
                if utils.wait_for_status(
                    volume_client.volumes.get,
                    volume.id,
                    success_status=['available'],
                    error_status=['error'],
                    sleep_time=1
                ):
                    volume_client.volumes.set_bootable(
                        volume.id,
                        parsed_args.bootable
                    )
                else:
                    msg = _(
                        "Volume status is not available for setting boot "
                        "state"
                    )
                    raise exceptions.CommandError(msg)
            except Exception as e:
                LOG.error(_("Failed to set volume bootable property: %s"), e)
        if parsed_args.read_only or parsed_args.read_write:
            try:
                if utils.wait_for_status(
                    volume_client.volumes.get,
                    volume.id,
                    success_status=['available'],
                    error_status=['error'],
                    sleep_time=1
                ):
                    volume_client.volumes.update_readonly_flag(
                        volume.id,
                        parsed_args.read_only
                    )
                else:
                    msg = _(
                        "Volume status is not available for setting it"
                        "read only."
                    )
                    raise exceptions.CommandError(msg)
            except Exception as e:
                LOG.error(_("Failed to set volume read-only access "
                            "mode flag: %s"), e)

        # Map 'metadata' column to 'properties'
        volume._info.update(
            {
                'properties':
                format_columns.DictColumn(volume._info.pop('metadata')),
                'type': volume._info.pop('volume_type'),
            },
        )
        # Replace "display_name" by "name", keep consistent in v1 and v2
        if 'display_name' in volume._info:
            volume._info.update({'name': volume._info.pop('display_name')})
        volume_info = utils.backward_compat_col_showone(
            volume._info, parsed_args.columns, {'display_name': 'name'}
        )

        return zip(*sorted(volume_info.items()))


class DeleteVolume(command.Command):
    _description = _("Delete volume(s)")

    def get_parser(self, prog_name):
        parser = super(DeleteVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volumes',
            metavar='<volume>',
            nargs="+",
            help=_('Volume(s) to delete (name or ID)'),
        )
        parser.add_argument(
            '--force',
            action='store_true',
            default=False,
            help=_('Attempt forced removal of volume(s), regardless of state '
                   '(defaults to False)'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        result = 0

        for i in parsed_args.volumes:
            try:
                volume_obj = utils.find_resource(
                    volume_client.volumes, i)
                if parsed_args.force:
                    volume_client.volumes.force_delete(volume_obj.id)
                else:
                    volume_client.volumes.delete(volume_obj.id)
            except Exception as e:
                result += 1
                LOG.error(_("Failed to delete volume with "
                            "name or ID '%(volume)s': %(e)s"),
                          {'volume': i, 'e': e})

        if result > 0:
            total = len(parsed_args.volumes)
            msg = (_("%(result)s of %(total)s volumes failed "
                   "to delete.") % {'result': result, 'total': total})
            raise exceptions.CommandError(msg)


class ListVolume(command.Lister):
    _description = _("List volumes")

    def get_parser(self, prog_name):
        parser = super(ListVolume, self).get_parser(prog_name)
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Filter results by volume name'),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            help=_('Filter results by status'),
        )
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=_('Include all projects (admin only)'),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        parser.add_argument(
            '--offset',
            type=int,
            action=parseractions.NonNegativeAction,
            metavar='<offset>',
            help=_('Index from which to start listing volumes'),
        )
        parser.add_argument(
            '--limit',
            type=int,
            action=parseractions.NonNegativeAction,
            metavar='<num-volumes>',
            help=_('Maximum number of volumes to display'),
        )
        return parser

    def take_action(self, parsed_args):

        volume_client = self.app.client_manager.volume
        compute_client = self.app.client_manager.compute

        if parsed_args.long:
            columns = (
                'ID',
                'Display Name',
                'Status',
                'Size',
                'Volume Type',
                'Bootable',
                'Attachments',
                'Metadata',
            )
            column_headers = (
                'ID',
                'Name',
                'Status',
                'Size',
                'Type',
                'Bootable',
                'Attached to',
                'Properties',
            )
        else:
            columns = (
                'ID',
                'Display Name',
                'Status',
                'Size',
                'Attachments',
            )
            column_headers = (
                'ID',
                'Name',
                'Status',
                'Size',
                'Attached to',
            )

        # Cache the server list
        server_cache = {}
        try:
            for s in compute_client.servers.list():
                server_cache[s.id] = s
        except Exception:
            # Just forget it if there's any trouble
            pass
        AttachmentsColumnWithCache = functools.partial(
            AttachmentsColumn, server_cache=server_cache)

        search_opts = {
            'all_tenants': parsed_args.all_projects,
            'display_name': parsed_args.name,
            'status': parsed_args.status,
        }

        if parsed_args.offset:
            search_opts['offset'] = parsed_args.offset

        data = volume_client.volumes.list(
            search_opts=search_opts,
            limit=parsed_args.limit,
        )
        column_headers = utils.backward_compat_col_lister(
            column_headers, parsed_args.columns, {'Display Name': 'Name'})

        return (column_headers,
                (utils.get_item_properties(
                    s, columns,
                    formatters={'Metadata': format_columns.DictColumn,
                                'Attachments': AttachmentsColumnWithCache},
                ) for s in data))


class MigrateVolume(command.Command):
    _description = _("Migrate volume to a new host")

    def get_parser(self, prog_name):
        parser = super(MigrateVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar="<volume>",
            help=_("Volume to migrate (name or ID)")
        )
        parser.add_argument(
            '--host',
            metavar="<host>",
            required=True,
            help=_("Destination host (takes the form: host@backend-name#pool)")
        )
        parser.add_argument(
            '--force-host-copy',
            action="store_true",
            help=_("Enable generic host-based force-migration, "
                   "which bypasses driver optimizations")
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        volume = utils.find_resource(volume_client.volumes, parsed_args.volume)
        volume_client.volumes.migrate_volume(volume.id, parsed_args.host,
                                             parsed_args.force_host_copy,)


class SetVolume(command.Command):
    _description = _("Set volume properties")

    def get_parser(self, prog_name):
        parser = super(SetVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=_('Volume to modify (name or ID)'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New volume name'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New volume description'),
        )
        parser.add_argument(
            '--size',
            metavar='<size>',
            type=int,
            help=_('Extend volume size in GB'),
        )
        parser.add_argument(
            "--no-property",
            dest="no_property",
            action="store_true",
            help=_("Remove all properties from <volume> "
                   "(specify both --no-property and --property to "
                   "remove the current properties before setting "
                   "new properties.)"),
        )
        parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            help=_('Set a property on this volume '
                   '(repeat option to set multiple properties)'),
        )
        bootable_group = parser.add_mutually_exclusive_group()
        bootable_group.add_argument(
            "--bootable",
            action="store_true",
            help=_("Mark volume as bootable")
        )
        bootable_group.add_argument(
            "--non-bootable",
            action="store_true",
            help=_("Mark volume as non-bootable")
        )
        readonly_group = parser.add_mutually_exclusive_group()
        readonly_group.add_argument(
            "--read-only",
            action="store_true",
            help=_("Set volume to read-only access mode")
        )
        readonly_group.add_argument(
            "--read-write",
            action="store_true",
            help=_("Set volume to read-write access mode")
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        volume = utils.find_resource(volume_client.volumes, parsed_args.volume)

        result = 0
        if parsed_args.size:
            try:
                if volume.status != 'available':
                    msg = (_("Volume is in %s state, it must be available "
                           "before size can be extended") % volume.status)
                    raise exceptions.CommandError(msg)
                if parsed_args.size <= volume.size:
                    msg = (_("New size must be greater than %s GB")
                           % volume.size)
                    raise exceptions.CommandError(msg)
                volume_client.volumes.extend(volume.id, parsed_args.size)
            except Exception as e:
                LOG.error(_("Failed to set volume size: %s"), e)
                result += 1

        if parsed_args.no_property:
            try:
                volume_client.volumes.delete_metadata(
                    volume.id, volume.metadata.keys())
            except Exception as e:
                LOG.error(_("Failed to clean volume properties: %s"), e)
                result += 1

        if parsed_args.property:
            try:
                volume_client.volumes.set_metadata(
                    volume.id,
                    parsed_args.property)
            except Exception as e:
                LOG.error(_("Failed to set volume property: %s"), e)
                result += 1
        if parsed_args.bootable or parsed_args.non_bootable:
            try:
                volume_client.volumes.set_bootable(
                    volume.id, parsed_args.bootable)
            except Exception as e:
                LOG.error(_("Failed to set volume bootable property: %s"), e)
                result += 1
        if parsed_args.read_only or parsed_args.read_write:
            try:
                volume_client.volumes.update_readonly_flag(
                    volume.id,
                    parsed_args.read_only)
            except Exception as e:
                LOG.error(_("Failed to set volume read-only access "
                            "mode flag: %s"), e)
                result += 1
        kwargs = {}
        if parsed_args.name:
            kwargs['display_name'] = parsed_args.name
        if parsed_args.description:
            kwargs['display_description'] = parsed_args.description
        if kwargs:
            try:
                volume_client.volumes.update(volume.id, **kwargs)
            except Exception as e:
                LOG.error(_("Failed to update volume display name "
                          "or display description: %s"), e)
                result += 1

        if result > 0:
            raise exceptions.CommandError(_("One or more of the "
                                          "set operations failed"))


class ShowVolume(command.ShowOne):
    _description = _("Show volume details")

    def get_parser(self, prog_name):
        parser = super(ShowVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=_('Volume to display (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        volume = utils.find_resource(volume_client.volumes, parsed_args.volume)
        # Map 'metadata' column to 'properties'
        volume._info.update(
            {
                'properties':
                format_columns.DictColumn(volume._info.pop('metadata')),
                'type': volume._info.pop('volume_type'),
            },
        )
        if 'os-vol-tenant-attr:tenant_id' in volume._info:
            volume._info.update(
                {'project_id': volume._info.pop(
                    'os-vol-tenant-attr:tenant_id')}
            )
        # Replace "display_name" by "name", keep consistent in v1 and v2
        if 'display_name' in volume._info:
            volume._info.update({'name': volume._info.pop('display_name')})

        volume_info = utils.backward_compat_col_showone(
            volume._info, parsed_args.columns, {'display_name': 'name'}
        )

        return zip(*sorted(volume_info.items()))


class UnsetVolume(command.Command):
    _description = _("Unset volume properties")

    def get_parser(self, prog_name):
        parser = super(UnsetVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=_('Volume to modify (name or ID)'),
        )
        parser.add_argument(
            '--property',
            metavar='<key>',
            action='append',
            help=_('Remove a property from volume '
                   '(repeat option to remove multiple properties)'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        volume = utils.find_resource(
            volume_client.volumes, parsed_args.volume)

        if parsed_args.property:
            volume_client.volumes.delete_metadata(
                volume.id,
                parsed_args.property,
            )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from cinderclient import api_versions
from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import volume_group_type


class TestVolumeGroupType(volume_fakes.TestVolume):

    def setUp(self):
        super().setUp()

        self.volume_group_types_mock = \
            self.app.client_manager.volume.group_types
        self.volume_group_types_mock.reset_mock()


class TestVolumeGroupTypeCreate(TestVolumeGroupType):

    maxDiff = 2000

    fake_volume_group_type = volume_fakes.create_one_volume_group_type()

    columns = (
        'ID',
        'Name',
        'Description',
        'Is Public',
        'Properties',
    )
    data = (
        fake_volume_group_type.id,
        fake_volume_group_type.name,
        fake_volume_group_type.description,
        fake_volume_group_type.is_public,
        format_columns.DictColumn(fake_volume_group_type.group_specs),
    )

    def setUp(self):
        super().setUp()

        self.volume_group_types_mock.create.return_value = \
            self.fake_volume_group_type

        self.cmd = volume_group_type.CreateVolumeGroupType(self.app, None)

    def test_volume_group_type_create(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.11')

        arglist = [
            self.fake_volume_group_type.name,
        ]
        verifylist = [
            ('name', self.fake_volume_group_type.name),
            ('description', None),
            ('is_public', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_group_types_mock.create.assert_called_once_with(
            self.fake_volume_group_type.name,
            None,
            True)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_type_create_with_options(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.11')

        arglist = [
            self.fake_volume_group_type.name,
            '--description', 'foo',
            '--private',
        ]
        verifylist = [
            ('name', self.fake_volume_group_type.name),
            ('description', 'foo'),
            ('is_public', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_group_types_mock.create.assert_called_once_with(
            self.fake_volume_group_type.name,
            'foo',
            False)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_type_create_pre_v311(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.10')

        arglist = [
            self.fake_volume_group_type.name,
        ]
        verifylist = [
            ('name', self.fake_volume_group_type.name),
            ('description', None),
            ('is_public', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.11 or greater is required',
            str(exc))


class TestVolumeGroupTypeDelete(TestVolumeGroupType):

    fake_volume_group_type = volume_fakes.create_one_volume_group_type()

    def setUp(self):
        super().setUp()

        self.volume_group_types_mock.get.return_value = \
            self.fake_volume_group_type
        self.volume_group_types_mock.delete.return_value = None

        self.cmd = volume_group_type.DeleteVolumeGroupType(self.app, None)

    def test_volume_group_type_delete(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.11')

        arglist = [
            self.fake_volume_group_type.id,
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_group_types_mock.delete.assert_called_once_with(
            self.fake_volume_group_type.id,
        )
        self.assertIsNone(result)

    def test_volume_group_type_delete_pre_v311(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.10')

        arglist = [
            self.fake_volume_group_type.id,
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.11 or greater is required',
            str(exc))


class TestVolumeGroupTypeSet(TestVolumeGroupType):

    fake_volume_group_type = volume_fakes.create_one_volume_group_type(
        methods={
            'get_keys': {'foo': 'bar'},
            'set_keys': None,
            'unset_keys': None,
        },
    )

    columns = (
        'ID',
        'Name',
        'Description',
        'Is Public',
        'Properties',
    )
    data = (
        fake_volume_group_type.id,
        fake_volume_group_type.name,
        fake_volume_group_type.description,
        fake_volume_group_type.is_public,
        format_columns.DictColumn(fake_volume_group_type.group_specs),
    )

    def setUp(self):
        super().setUp()

        self.volume_group_types_mock.get.return_value = \
            self.fake_volume_group_type
        self.volume_group_types_mock.update.return_value = \
            self.fake_volume_group_type

        self.cmd = volume_group_type.SetVolumeGroupType(self.app, None)

    def test_volume_group_type_set(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.11')

        self.fake_volume_group_type.set_keys.return_value = None

        arglist = [
            self.fake_volume_group_type.id,
            '--name', 'foo',
            '--description', 'hello, world',
            '--public',
            '--property', 'fizz=buzz',
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
            ('name', 'foo'),
            ('description', 'hello, world'),
            ('is_public', True),
            ('no_property', False),
            ('properties', {'fizz': 'buzz'}),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_group_types_mock.update.assert_called_once_with(
            self.fake_volume_group_type.id,
            name='foo',
            description='hello, world',
            is_public=True,
        )
        self.fake_volume_group_type.set_keys.assert_called_once_with(
            {'fizz': 'buzz'},
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_type_with_no_property_option(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.11')

        arglist = [
            self.fake_volume_group_type.id,
            '--no-property',
            '--property', 'fizz=buzz',
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
            ('name', None),
            ('description', None),
            ('is_public', None),
            ('no_property', True),
            ('properties', {'fizz': 'buzz'}),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_group_types_mock.get.assert_called_once_with(
            self.fake_volume_group_type.id)
        self.fake_volume_group_type.get_keys.assert_called_once_with()
        self.fake_volume_group_type.unset_keys.assert_called_once_with(
            {'foo': 'bar'}.keys())
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_type_set_pre_v311(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.10')

        arglist = [
            self.fake_volume_group_type.id,
            '--name', 'foo',
            '--description', 'hello, world',
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
            ('name', 'foo'),
            ('description', 'hello, world'),
            ('is_public', None),
            ('no_property', False),
            ('properties', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.11 or greater is required',
            str(exc))


class TestVolumeGroupTypeUnset(TestVolumeGroupType):

    fake_volume_group_type = volume_fakes.create_one_volume_group_type(
        methods={'unset_keys': None},
    )

    columns = (
        'ID',
        'Name',
        'Description',
        'Is Public',
        'Properties',
    )
    data = (
        fake_volume_group_type.id,
        fake_volume_group_type.name,
        fake_volume_group_type.description,
        fake_volume_group_type.is_public,
        format_columns.DictColumn(fake_volume_group_type.group_specs),
    )

    def setUp(self):
        super().setUp()

        self.volume_group_types_mock.get.return_value = \
            self.fake_volume_group_type

        self.cmd = volume_group_type.UnsetVolumeGroupType(self.app, None)

    def test_volume_group_type_unset(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.11')

        arglist = [
            self.fake_volume_group_type.id,
            '--property', 'fizz',
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
            ('properties', ['fizz']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_group_types_mock.get.assert_has_calls([
            mock.call(self.fake_volume_group_type.id),
            mock.call(self.fake_volume_group_type.id),
        ])
        self.fake_volume_group_type.unset_keys.assert_called_once_with(
            ['fizz'])
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_group_type_unset_pre_v311(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.10')

        arglist = [
            self.fake_volume_group_type.id,
            '--property', 'fizz',
        ]
        verifylist = [
            ('group_type', self.fake_volume_group_type.id),
            ('properties', ['fizz']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.11 or greater is required',
            str(exc))


class TestVolumeGroupTypeList(TestVolumeGroupType):

    fake_volume_group_types = volume_fakes.create_volume_group_types()

    columns = (
        'ID',
        'Name',
        'Is Public',
        'Properties',
    )
    data = [
        (
            fake_volume_group_type.id,
            fake_volume_group_type.name,
            fake_volume_group_type.is_public,
            fake_volume_group_type.group_specs,
        ) for fake_volume_group_type in fake_volume_group_types
    ]

    def setUp(self):
        super().setUp()

        self.volume_group_types_mock.list.return_value = \
            self.fake_volume_group_types
        self.volume_group_types_mock.default.return_value = \
            self.fake_volume_group_types[0]

        self.cmd = volume_group_type.ListVolumeGroupType(self.app, None)

    def test_volume_group_type_list(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.11')

        arglist = [
        ]
        verifylist = [
            ('show_default', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_group_types_mock.list.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(tuple(self.data), data)

    def test_volume_group_type_list_with_default_option(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.11')

        arglist = [
            '--default',
        ]
        verifylist = [
            ('show_default', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_group_types_mock.default.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(tuple([self.data[0]]), data)

    def test_volume_group_type_list_pre_v311(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.10')

        arglist = [
        ]
        verifylist = [
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.11 or greater is required',
            str(exc))



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from openstackclient.image.v2 import metadef_namespaces
from openstackclient.tests.unit.image.v2 import fakes as md_namespace_fakes


class TestMetadefNamespaces(md_namespace_fakes.TestMetadefNamespaces):
    def setUp(self):
        super().setUp()

        # Get shortcuts to mocked image client
        self.client = self.app.client_manager.image

        # Get shortcut to the Mocks in identity client
        self.project_mock = self.app.client_manager.identity.projects
        self.project_mock.reset_mock()
        self.domain_mock = self.app.client_manager.identity.domains
        self.domain_mock.reset_mock()


class TestMetadefNamespaceCreate(TestMetadefNamespaces):
    _metadef_namespace = md_namespace_fakes.create_one_metadef_namespace()

    expected_columns = (
        'created_at',
        'description',
        'display_name',
        'id',
        'is_protected',
        'location',
        'name',
        'namespace',
        'owner',
        'resource_type_associations',
        'updated_at',
        'visibility'
    )
    expected_data = (
        _metadef_namespace.created_at,
        _metadef_namespace.description,
        _metadef_namespace.display_name,
        _metadef_namespace.id,
        _metadef_namespace.is_protected,
        _metadef_namespace.location,
        _metadef_namespace.name,
        _metadef_namespace.namespace,
        _metadef_namespace.owner,
        _metadef_namespace.resource_type_associations,
        _metadef_namespace.updated_at,
        _metadef_namespace.visibility
    )

    def setUp(self):
        super().setUp()

        self.client.create_metadef_namespace.return_value \
            = self._metadef_namespace
        self.cmd = metadef_namespaces.CreateMetadefNameSpace(self.app, None)
        self.datalist = self._metadef_namespace

    def test_namespace_create(self):
        arglist = [
            self._metadef_namespace.namespace
        ]

        verifylist = [

        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)


class TestMetadefNamespaceDelete(TestMetadefNamespaces):
    _metadef_namespace = md_namespace_fakes.create_one_metadef_namespace()

    def setUp(self):
        super().setUp()

        self.client.delete_metadef_namespace.return_value \
            = self._metadef_namespace
        self.cmd = metadef_namespaces.DeleteMetadefNameSpace(self.app, None)
        self.datalist = self._metadef_namespace

    def test_namespace_create(self):
        arglist = [
            self._metadef_namespace.namespace
        ]

        verifylist = [

        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)


class TestMetadefNamespaceList(TestMetadefNamespaces):
    _metadef_namespace = [md_namespace_fakes.create_one_metadef_namespace()]

    columns = [
        'namespace'
    ]

    datalist = []

    def setUp(self):
        super().setUp()

        self.client.metadef_namespaces.side_effect = [
            self._metadef_namespace, []]

        # Get the command object to test
        self.client.metadef_namespaces.return_value = iter(
            self._metadef_namespace
        )
        self.cmd = metadef_namespaces.ListMetadefNameSpaces(self.app, None)
        self.datalist = self._metadef_namespace

    def test_namespace_list_no_options(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        self.assertEqual(getattr(self.datalist[0], 'namespace'),
                         next(data)[0])


class TestMetadefNamespaceSet(TestMetadefNamespaces):
    _metadef_namespace = md_namespace_fakes.create_one_metadef_namespace()

    def setUp(self):
        super().setUp()

        self.client.update_metadef_namespace.return_value \
            = self._metadef_namespace
        self.cmd = metadef_namespaces.SetMetadefNameSpace(self.app, None)
        self.datalist = self._metadef_namespace

    def test_namespace_set_no_options(self):
        arglist = [
            self._metadef_namespace.namespace
        ]
        verifylist = [
            ('namespace', self._metadef_namespace.namespace),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.assertIsNone(result)


class TestMetadefNamespaceShow(TestMetadefNamespaces):
    _metadef_namespace = md_namespace_fakes.create_one_metadef_namespace()

    expected_columns = (
        'created_at',
        'display_name',
        'namespace',
        'owner',
        'visibility'
    )
    expected_data = (
        _metadef_namespace.created_at,
        _metadef_namespace.display_name,
        _metadef_namespace.namespace,
        _metadef_namespace.owner,
        _metadef_namespace.visibility
    )

    def setUp(self):
        super().setUp()

        self.client.get_metadef_namespace.return_value \
            = self._metadef_namespace
        self.cmd = metadef_namespaces.ShowMetadefNameSpace(self.app, None)

    def test_namespace_show_no_options(self):
        arglist = [
            self._metadef_namespace.namespace
        ]

        verifylist = [

        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.expected_columns, columns)
        self.assertEqual(self.expected_data, data)



# Copyright (c) 2017, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib import exceptions as tempest_exc

from openstackclient.tests.functional import base


class ExtensionTests(base.TestCase):
    """Functional tests for extension"""

    @classmethod
    def setUpClass(cls):
        super(ExtensionTests, cls).setUpClass()
        cls.haz_network = cls.is_service_enabled('network')

    def test_extension_list_compute(self):
        """Test compute extension list"""
        output = self.openstack(
            'extension list --compute',
            parse_output=True,
        )
        name_list = [item.get('Name') for item in output]
        self.assertIn(
            'ImageSize',
            name_list,
        )

    def test_extension_list_volume(self):
        """Test volume extension list"""
        output = self.openstack(
            'extension list --volume',
            parse_output=True,
        )
        name_list = [item.get('Name') for item in output]
        self.assertIn(
            'TypesManage',
            name_list,
        )

    def test_extension_list_network(self):
        """Test network extension list"""
        if not self.haz_network:
            self.skipTest("No Network service present")

        output = self.openstack(
            'extension list --network',
            parse_output=True,
        )
        name_list = [item.get('Name') for item in output]
        self.assertIn(
            'Default Subnetpools',
            name_list,
        )

    def test_extension_show_network(self):
        """Test network extension show"""
        if not self.haz_network:
            self.skipTest("No Network service present")

        name = 'agent'
        output = self.openstack(
            'extension show ' + name,
            parse_output=True,
        )
        self.assertEqual(
            name,
            output.get('alias'),
        )

    def test_extension_show_not_exist(self):
        """Test extension show with not existed name"""
        if not self.haz_network:
            self.skipTest("No Network service present")

        name = 'not_existed_ext'
        try:
            self.openstack('extension show ' + name)
        except tempest_exc.CommandFailed as e:
            self.assertIn('No Extension found for', str(e))
            self.assertIn(name, str(e))
        else:
            self.fail('CommandFailed should be raised')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from osc_lib.cli import format_columns

from openstackclient.image.v2 import task
from openstackclient.tests.unit.image.v2 import fakes as image_fakes


class TestTask(image_fakes.TestImagev2):
    def setUp(self):
        super().setUp()

        # Get shortcuts to mocked image client
        self.client = self.app.client_manager.image


class TestTaskShow(TestTask):

    task = image_fakes.create_one_task()

    columns = (
        'created_at',
        'expires_at',
        'id',
        'input',
        'message',
        'owner_id',
        'properties',
        'result',
        'status',
        'type',
        'updated_at',
    )
    data = (
        task.created_at,
        task.expires_at,
        task.id,
        task.input,
        task.message,
        task.owner_id,
        format_columns.DictColumn({}),
        task.result,
        task.status,
        task.type,
        task.updated_at,
    )

    def setUp(self):
        super().setUp()

        self.client.get_task.return_value = self.task

        # Get the command object to test
        self.cmd = task.ShowTask(self.app, None)

    def test_task_show(self):
        arglist = [self.task.id]
        verifylist = [
            ('task', self.task.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class ShowOne in cliff, abstract method take_action()
        # returns a two-part tuple with a tuple of column names and a tuple of
        # data to be shown.
        columns, data = self.cmd.take_action(parsed_args)
        self.client.get_task.assert_called_with(self.task.id)

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestTaskList(TestTask):

    tasks = image_fakes.create_tasks()

    columns = (
        'ID',
        'Type',
        'Status',
        'Owner',
    )
    datalist = [
        (
            task.id,
            task.type,
            task.status,
            task.owner_id,
        )
        for task in tasks
    ]

    def setUp(self):
        super().setUp()

        self.client.tasks.side_effect = [self.tasks, []]

        # Get the command object to test
        self.cmd = task.ListTask(self.app, None)

    def test_task_list_no_options(self):
        arglist = []
        verifylist = [
            ('sort_key', None),
            ('sort_dir', None),
            ('limit', None),
            ('marker', None),
            ('type', None),
            ('status', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.tasks.assert_called_with()

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, data)

    def test_task_list_sort_key_option(self):
        arglist = ['--sort-key', 'created_at']
        verifylist = [('sort_key', 'created_at')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.client.tasks.assert_called_with(
            sort_key=parsed_args.sort_key,
        )

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.datalist, data)

    def test_task_list_sort_dir_option(self):
        arglist = ['--sort-dir', 'desc']
        verifylist = [('sort_dir', 'desc')]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.client.tasks.assert_called_with(
            sort_dir=parsed_args.sort_dir,
        )

    def test_task_list_pagination_options(self):
        arglist = ['--limit', '1', '--marker', self.tasks[0].id]
        verifylist = [('limit', 1), ('marker', self.tasks[0].id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.client.tasks.assert_called_with(
            limit=parsed_args.limit,
            marker=parsed_args.marker,
        )

    def test_task_list_type_option(self):
        arglist = ['--type', self.tasks[0].type]
        verifylist = [('type', self.tasks[0].type)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.client.tasks.assert_called_with(
            type=self.tasks[0].type,
        )

    def test_task_list_status_option(self):
        arglist = ['--status', self.tasks[0].status]
        verifylist = [('status', self.tasks[0].status)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        self.client.tasks.assert_called_with(
            status=self.tasks[0].status,
        )



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from cinderclient import api_versions
import ddt
from osc_lib import exceptions

from openstackclient.tests.unit import utils as tests_utils
from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import block_storage_log_level as service


class TestService(volume_fakes.TestVolume):

    def setUp(self):
        super().setUp()

        # Get a shortcut to the ServiceManager Mock
        self.service_mock = self.app.client_manager.volume.services
        self.service_mock.reset_mock()


class TestBlockStorageLogLevelList(TestService):

    service_log = volume_fakes.create_service_log_level_entry()

    def setUp(self):
        super().setUp()

        self.service_mock.get_log_levels.return_value = [self.service_log]

        # Get the command object to test
        self.cmd = service.BlockStorageLogLevelList(self.app, None)

    def test_block_storage_log_level_list(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.32')
        arglist = [
            '--host', self.service_log.host,
            '--service', self.service_log.binary,
            '--log-prefix', self.service_log.prefix,
        ]
        verifylist = [
            ('host', self.service_log.host),
            ('service', self.service_log.binary),
            ('log_prefix', self.service_log.prefix),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        expected_columns = [
            'Binary',
            'Host',
            'Prefix',
            'Level',
        ]

        # confirming if all expected columns are present in the result.
        self.assertEqual(expected_columns, columns)

        datalist = ((
            self.service_log.binary,
            self.service_log.host,
            self.service_log.prefix,
            self.service_log.level,
        ), )

        # confirming if all expected values are present in the result.
        self.assertEqual(datalist, tuple(data))

        # checking if proper call was made to get log level of services
        self.service_mock.get_log_levels.assert_called_with(
            server=self.service_log.host,
            binary=self.service_log.binary,
            prefix=self.service_log.prefix,
        )

    def test_block_storage_log_level_list_pre_332(self):
        arglist = [
            '--host', self.service_log.host,
            '--service', 'cinder-api',
            '--log-prefix', 'cinder_test.api.common',
        ]
        verifylist = [
            ('host', self.service_log.host),
            ('service', 'cinder-api'),
            ('log_prefix', 'cinder_test.api.common'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(exceptions.CommandError, self.cmd.take_action,
                                parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.32 or greater is required', str(exc))

    def test_block_storage_log_level_list_invalid_service_name(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.32')
        arglist = [
            '--host', self.service_log.host,
            '--service', 'nova-api',
            '--log-prefix', 'cinder_test.api.common',
        ]
        verifylist = [
            ('host', self.service_log.host),
            ('service', 'nova-api'),
            ('log_prefix', 'cinder_test.api.common'),
        ]

        self.assertRaises(tests_utils.ParserException, self.check_parser,
                          self.cmd, arglist, verifylist)


@ddt.ddt
class TestBlockStorageLogLevelSet(TestService):

    service_log = volume_fakes.create_service_log_level_entry()

    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = service.BlockStorageLogLevelSet(self.app, None)

    def test_block_storage_log_level_set(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.32')
        arglist = [
            'ERROR',
            '--host', self.service_log.host,
            '--service', self.service_log.binary,
            '--log-prefix', self.service_log.prefix,
        ]
        verifylist = [
            ('level', 'ERROR'),
            ('host', self.service_log.host),
            ('service', self.service_log.binary),
            ('log_prefix', self.service_log.prefix),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # checking if proper call was made to set log level of services
        self.service_mock.set_log_levels.assert_called_with(
            level='ERROR',
            server=self.service_log.host,
            binary=self.service_log.binary,
            prefix=self.service_log.prefix,
        )

    def test_block_storage_log_level_set_pre_332(self):
        arglist = [
            'ERROR',
            '--host', self.service_log.host,
            '--service', 'cinder-api',
            '--log-prefix', 'cinder_test.api.common',
        ]
        verifylist = [
            ('level', 'ERROR'),
            ('host', self.service_log.host),
            ('service', 'cinder-api'),
            ('log_prefix', 'cinder_test.api.common'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(exceptions.CommandError, self.cmd.take_action,
                                parsed_args)
        self.assertIn(
            '--os-volume-api-version 3.32 or greater is required', str(exc))

    def test_block_storage_log_level_set_invalid_service_name(self):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.32')
        arglist = [
            'ERROR',
            '--host', self.service_log.host,
            '--service', 'nova-api',
            '--log-prefix', 'cinder.api.common',
        ]
        verifylist = [
            ('level', 'ERROR'),
            ('host', self.service_log.host),
            ('service', 'nova-api'),
            ('log_prefix', 'cinder.api.common'),
        ]

        self.assertRaises(tests_utils.ParserException, self.check_parser,
                          self.cmd, arglist, verifylist)

    @ddt.data('WARNING', 'info', 'Error', 'debuG', 'fake-log-level')
    def test_block_storage_log_level_set_log_level(self, log_level):
        self.app.client_manager.volume.api_version = \
            api_versions.APIVersion('3.32')
        arglist = [
            log_level,
            '--host', self.service_log.host,
            '--service', 'cinder-api',
            '--log-prefix', 'cinder.api.common',
        ]
        verifylist = [
            ('level', log_level.upper()),
            ('host', self.service_log.host),
            ('service', 'cinder-api'),
            ('log_prefix', 'cinder.api.common'),
        ]

        if log_level == 'fake-log-level':
            self.assertRaises(tests_utils.ParserException, self.check_parser,
                              self.cmd, arglist, verifylist)
        else:
            parsed_args = self.check_parser(self.cmd, arglist, verifylist)

            self.cmd.take_action(parsed_args)

            # checking if proper call was made to set log level of services
            self.service_mock.set_log_levels.assert_called_with(
                level=log_level.upper(),
                server=self.service_log.host,
                binary=self.service_log.binary,
                prefix=self.service_log.prefix)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import uuid

from openstackclient.tests.functional.network.v2 import common


class NetworkTrunkTests(common.NetworkTests):
    """Functional tests for Network Trunks"""

    def setUp(self):
        super().setUp()
        # Nothing in this class works with Nova Network
        if not self.haz_network:
            self.skipTest("No Network service present")

        network_name = uuid.uuid4().hex
        subnet_name = uuid.uuid4().hex
        self.parent_port_name = uuid.uuid4().hex
        self.sub_port_name = uuid.uuid4().hex

        self.openstack('network create %s' % network_name)
        self.addCleanup(self.openstack, 'network delete %s' % network_name)

        self.openstack(
            'subnet create %s '
            '--network %s --subnet-range 10.0.0.0/24' % (
                subnet_name, network_name))
        self.openstack('port create %s --network %s' %
                       (self.parent_port_name, network_name))
        self.addCleanup(self.openstack, 'port delete %s' %
                        self.parent_port_name)
        json_out = self.openstack('port create %s --network %s -f json' %
                                  (self.sub_port_name, network_name))
        self.sub_port_id = json.loads(json_out)['id']
        self.addCleanup(self.openstack, 'port delete %s' % self.sub_port_name)

    def test_network_trunk_create_delete(self):
        trunk_name = uuid.uuid4().hex
        self.openstack('network trunk create %s --parent-port %s -f json ' %
                       (trunk_name, self.parent_port_name))
        raw_output = self.openstack(
            'network trunk delete ' +
            trunk_name
        )
        self.assertEqual('', raw_output)

    def test_network_trunk_list(self):
        trunk_name = uuid.uuid4().hex
        json_output = json.loads(self.openstack(
            'network trunk create %s --parent-port %s -f json ' %
            (trunk_name, self.parent_port_name)))
        self.addCleanup(self.openstack,
                        'network trunk delete ' + trunk_name)
        self.assertEqual(trunk_name, json_output['name'])

        json_output = json.loads(self.openstack(
            'network trunk list -f json'
        ))
        self.assertIn(trunk_name, [tr['Name'] for tr in json_output])

    def test_network_trunk_set_unset(self):
        trunk_name = uuid.uuid4().hex
        json_output = json.loads(self.openstack(
            'network trunk create %s --parent-port %s -f json ' %
            (trunk_name, self.parent_port_name)))
        self.addCleanup(self.openstack,
                        'network trunk delete ' + trunk_name)
        self.assertEqual(trunk_name, json_output['name'])

        self.openstack(
            'network trunk set '
            '--enable ' +
            trunk_name
        )

        json_output = json.loads(self.openstack(
            'network trunk show -f json ' +
            trunk_name
        ))
        self.assertTrue(json_output['is_admin_state_up'])

        # Add subport to trunk
        self.openstack(
            'network trunk set ' +
            '--subport port=%s,segmentation-type=vlan,segmentation-id=42 ' %
            (self.sub_port_name) +
            trunk_name
        )
        json_output = json.loads(self.openstack(
            'network trunk show -f json ' +
            trunk_name
        ))
        self.assertEqual(
            [{
                'port_id': self.sub_port_id,
                'segmentation_id': 42,
                'segmentation_type': 'vlan'
            }],
            json_output['sub_ports'])

        # Remove subport from trunk
        self.openstack(
            'network trunk unset ' +
            trunk_name +
            ' --subport ' +
            self.sub_port_name
        )
        json_output = json.loads(self.openstack(
            'network trunk show -f json ' +
            trunk_name
        ))
        self.assertEqual(
            [],
            json_output['sub_ports'])

    def test_network_trunk_list_subports(self):
        trunk_name = uuid.uuid4().hex
        json_output = json.loads(self.openstack(
            'network trunk create %s --parent-port %s '
            '--subport port=%s,segmentation-type=vlan,segmentation-id=42 '
            '-f json ' %
            (trunk_name, self.parent_port_name, self.sub_port_name)))
        self.addCleanup(self.openstack,
                        'network trunk delete ' + trunk_name)
        self.assertEqual(trunk_name, json_output['name'])

        json_output = json.loads(self.openstack(
            'network subport list --trunk %s -f json' % trunk_name))
        self.assertEqual(
            [{
                'Port': self.sub_port_id,
                'Segmentation ID': 42,
                'Segmentation Type': 'vlan'
            }],
            json_output)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock
from unittest.mock import call

from osc_lib import exceptions

from openstackclient.network.v2 import network_segment
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestNetworkSegment(network_fakes.TestNetworkV2):

    def setUp(self):
        super(TestNetworkSegment, self).setUp()

        # Get a shortcut to the network client
        self.network = self.app.client_manager.network


class TestCreateNetworkSegment(TestNetworkSegment):

    # The network segment to create along with associated network.
    _network_segment = network_fakes.create_one_network_segment()
    _network = network_fakes.create_one_network({
        'id': _network_segment.network_id,
    })

    columns = (
        'description',
        'id',
        'name',
        'network_id',
        'network_type',
        'physical_network',
        'segmentation_id',
    )

    data = (
        _network_segment.description,
        _network_segment.id,
        _network_segment.name,
        _network_segment.network_id,
        _network_segment.network_type,
        _network_segment.physical_network,
        _network_segment.segmentation_id,
    )

    def setUp(self):
        super(TestCreateNetworkSegment, self).setUp()

        self.network.create_segment = mock.Mock(
            return_value=self._network_segment
        )
        self.network.find_network = mock.Mock(return_value=self._network)

        # Get the command object to test
        self.cmd = network_segment.CreateNetworkSegment(
            self.app,
            self.namespace
        )

    def test_create_no_options(self):
        # Missing required args should bail here
        self.assertRaises(tests_utils.ParserException, self.check_parser,
                          self.cmd, [], [])

    def test_create_invalid_network_type(self):
        arglist = [
            '--network', self._network_segment.network_id,
            '--network-type', 'foo',
            self._network_segment.name,
        ]
        self.assertRaises(tests_utils.ParserException, self.check_parser,
                          self.cmd, arglist, [])

    def test_create_minimum_options(self):
        arglist = [
            '--network', self._network_segment.network_id,
            '--network-type', self._network_segment.network_type,
            self._network_segment.name,
        ]
        verifylist = [
            ('network', self._network_segment.network_id),
            ('network_type', self._network_segment.network_type),
            ('name', self._network_segment.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network.find_network.assert_called_once_with(
            self._network_segment.network_id,
            ignore_missing=False
        )
        self.network.create_segment.assert_called_once_with(**{
            'network_id': self._network_segment.network_id,
            'network_type': self._network_segment.network_type,
            'name': self._network_segment.name,
        })

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_all_options(self):
        arglist = [
            '--description', self._network_segment.description,
            '--network', self._network_segment.network_id,
            '--network-type', self._network_segment.network_type,
            '--physical-network', self._network_segment.physical_network,
            '--segment', str(self._network_segment.segmentation_id),
            self._network_segment.name,
        ]
        verifylist = [
            ('description', self._network_segment.description),
            ('network', self._network_segment.network_id),
            ('network_type', self._network_segment.network_type),
            ('physical_network', self._network_segment.physical_network),
            ('segment', self._network_segment.segmentation_id),
            ('name', self._network_segment.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network.find_network.assert_called_once_with(
            self._network_segment.network_id,
            ignore_missing=False
        )
        self.network.create_segment.assert_called_once_with(**{
            'description': self._network_segment.description,
            'network_id': self._network_segment.network_id,
            'network_type': self._network_segment.network_type,
            'physical_network': self._network_segment.physical_network,
            'segmentation_id': self._network_segment.segmentation_id,
            'name': self._network_segment.name,
        })

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestDeleteNetworkSegment(TestNetworkSegment):

    # The network segments to delete.
    _network_segments = network_fakes.create_network_segments()

    def setUp(self):
        super(TestDeleteNetworkSegment, self).setUp()

        self.network.delete_segment = mock.Mock(return_value=None)
        self.network.find_segment = mock.Mock(
            side_effect=self._network_segments
        )

        # Get the command object to test
        self.cmd = network_segment.DeleteNetworkSegment(
            self.app,
            self.namespace
        )

    def test_delete(self):
        arglist = [
            self._network_segments[0].id,
        ]
        verifylist = [
            ('network_segment', [self._network_segments[0].id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network.delete_segment.assert_called_once_with(
            self._network_segments[0]
        )
        self.assertIsNone(result)

    def test_delete_multiple(self):
        arglist = []
        for _network_segment in self._network_segments:
            arglist.append(_network_segment.id)
        verifylist = [
            ('network_segment', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for _network_segment in self._network_segments:
            calls.append(call(_network_segment))
        self.network.delete_segment.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_delete_multiple_with_exception(self):
        arglist = [
            self._network_segments[0].id,
            'doesnotexist'
        ]
        verifylist = [
            ('network_segment', [self._network_segments[0].id,
                                 'doesnotexist']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [self._network_segments[0],
                            exceptions.CommandError]
        self.network.find_segment = (
            mock.Mock(side_effect=find_mock_result)
        )

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 network segments failed to delete.',
                             str(e))

        self.network.find_segment.assert_any_call(
            self._network_segments[0].id, ignore_missing=False)
        self.network.find_segment.assert_any_call(
            'doesnotexist', ignore_missing=False)
        self.network.delete_segment.assert_called_once_with(
            self._network_segments[0]
        )


class TestListNetworkSegment(TestNetworkSegment):
    _network = network_fakes.create_one_network()
    _network_segments = network_fakes.create_network_segments(count=3)

    columns = (
        'ID',
        'Name',
        'Network',
        'Network Type',
        'Segment',
    )
    columns_long = columns + (
        'Physical Network',
    )

    data = []
    for _network_segment in _network_segments:
        data.append((
            _network_segment.id,
            _network_segment.name,
            _network_segment.network_id,
            _network_segment.network_type,
            _network_segment.segmentation_id,
        ))

    data_long = []
    for _network_segment in _network_segments:
        data_long.append((
            _network_segment.id,
            _network_segment.name,
            _network_segment.network_id,
            _network_segment.network_type,
            _network_segment.segmentation_id,
            _network_segment.physical_network,
        ))

    def setUp(self):
        super(TestListNetworkSegment, self).setUp()

        # Get the command object to test
        self.cmd = network_segment.ListNetworkSegment(self.app, self.namespace)

        self.network.find_network = mock.Mock(return_value=self._network)
        self.network.segments = mock.Mock(return_value=self._network_segments)

    def test_list_no_option(self):
        arglist = []
        verifylist = [
            ('long', False),
            ('network', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network.segments.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
            ('network', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network.segments.assert_called_once_with()
        self.assertEqual(self.columns_long, columns)
        self.assertEqual(self.data_long, list(data))

    def test_list_network(self):
        arglist = [
            '--network',
            self._network.id,
        ]
        verifylist = [
            ('long', False),
            ('network', self._network.id)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network.segments.assert_called_once_with(
            **{'network_id': self._network.id}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


class TestSetNetworkSegment(TestNetworkSegment):

    # The network segment to show.
    _network_segment = network_fakes.create_one_network_segment()

    def setUp(self):
        super(TestSetNetworkSegment, self).setUp()

        self.network.find_segment = mock.Mock(
            return_value=self._network_segment
        )
        self.network.update_segment = mock.Mock(
            return_value=self._network_segment
        )

        # Get the command object to test
        self.cmd = network_segment.SetNetworkSegment(self.app, self.namespace)

    def test_set_no_options(self):
        arglist = [
            self._network_segment.id,
        ]
        verifylist = [
            ('network_segment', self._network_segment.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.network.update_segment.assert_called_once_with(
            self._network_segment, **{}
        )
        self.assertIsNone(result)

    def test_set_all_options(self):
        arglist = [
            '--description', 'new description',
            '--name', 'new name',
            self._network_segment.id,
        ]
        verifylist = [
            ('description', 'new description'),
            ('name', 'new name'),
            ('network_segment', self._network_segment.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'description': 'new description',
            'name': 'new name',
        }
        self.network.update_segment.assert_called_once_with(
            self._network_segment, **attrs
        )
        self.assertIsNone(result)


class TestShowNetworkSegment(TestNetworkSegment):

    # The network segment to show.
    _network_segment = network_fakes.create_one_network_segment()

    columns = (
        'description',
        'id',
        'name',
        'network_id',
        'network_type',
        'physical_network',
        'segmentation_id',
    )

    data = (
        _network_segment.description,
        _network_segment.id,
        _network_segment.name,
        _network_segment.network_id,
        _network_segment.network_type,
        _network_segment.physical_network,
        _network_segment.segmentation_id,
    )

    def setUp(self):
        super(TestShowNetworkSegment, self).setUp()

        self.network.find_segment = mock.Mock(
            return_value=self._network_segment
        )

        # Get the command object to test
        self.cmd = network_segment.ShowNetworkSegment(self.app, self.namespace)

    def test_show_no_options(self):
        # Missing required args should bail here
        self.assertRaises(tests_utils.ParserException, self.check_parser,
                          self.cmd, [], [])

    def test_show_all_options(self):
        arglist = [
            self._network_segment.id,
        ]
        verifylist = [
            ('network_segment', self._network_segment.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network.find_segment.assert_called_once_with(
            self._network_segment.id,
            ignore_missing=False
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Object Store v1 API Library Tests"""

from unittest import mock

from keystoneauth1 import session
from requests_mock.contrib import fixture

from openstackclient.api import object_store_v1 as object_store
from openstackclient.tests.unit import utils


FAKE_ACCOUNT = 'q12we34r'
FAKE_AUTH = '11223344556677889900'
FAKE_URL = 'http://gopher.com/v1/' + FAKE_ACCOUNT

FAKE_CONTAINER = 'rainbarrel'
FAKE_OBJECT = 'spigot'

LIST_CONTAINER_RESP = [
    {"name": "qaz", "count": 0, "bytes": 0,
     "last_modified": "2020-05-16T05:52:07.377550"},
    {"name": "fred", "count": 0, "bytes": 0,
     "last_modified": "2020-05-16T05:55:07.377550"},
]

LIST_OBJECT_RESP = [
    {'name': 'fred', 'bytes': 1234, 'content_type': 'text'},
    {'name': 'wilma', 'bytes': 5678, 'content_type': 'text'},
]


class TestObjectAPIv1(utils.TestCase):

    def setUp(self):
        super(TestObjectAPIv1, self).setUp()
        sess = session.Session()
        self.api = object_store.APIv1(session=sess, endpoint=FAKE_URL)
        self.requests_mock = self.useFixture(fixture.Fixture())


class TestContainer(TestObjectAPIv1):

    def setUp(self):
        super(TestContainer, self).setUp()

    def test_container_create(self):
        headers = {
            'x-trans-id': '1qaz2wsx',
        }
        self.requests_mock.register_uri(
            'PUT',
            FAKE_URL + '/qaz',
            headers=headers,
            status_code=201,
        )
        ret = self.api.container_create(container='qaz')
        data = {
            'account': FAKE_ACCOUNT,
            'container': 'qaz',
            'x-trans-id': '1qaz2wsx',
        }
        self.assertEqual(data, ret)

    def test_container_delete(self):
        self.requests_mock.register_uri(
            'DELETE',
            FAKE_URL + '/qaz',
            status_code=204,
        )
        ret = self.api.container_delete(container='qaz')
        self.assertIsNone(ret)

    def test_container_list_no_options(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL,
            json=LIST_CONTAINER_RESP,
            status_code=200,
        )
        ret = self.api.container_list()
        self.assertEqual(LIST_CONTAINER_RESP, ret)

    def test_container_list_prefix(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '?prefix=foo%2f&format=json',
            json=LIST_CONTAINER_RESP,
            status_code=200,
        )
        ret = self.api.container_list(
            prefix='foo/',
        )
        self.assertEqual(LIST_CONTAINER_RESP, ret)

    def test_container_list_marker_limit_end(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '?marker=next&limit=2&end_marker=stop&format=json',
            json=LIST_CONTAINER_RESP,
            status_code=200,
        )
        ret = self.api.container_list(
            marker='next',
            limit=2,
            end_marker='stop',
        )
        self.assertEqual(LIST_CONTAINER_RESP, ret)

    def test_container_list_full_listing(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '?limit=1&format=json',
            json=[LIST_CONTAINER_RESP[0]],
            status_code=200,
        )
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL +
            '?marker=%s&limit=1&format=json' % LIST_CONTAINER_RESP[0]['name'],
            json=[LIST_CONTAINER_RESP[1]],
            status_code=200,
        )
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL +
            '?marker=%s&limit=1&format=json' % LIST_CONTAINER_RESP[1]['name'],
            json=[],
            status_code=200,
        )
        ret = self.api.container_list(
            limit=1,
            full_listing=True,
        )
        self.assertEqual(LIST_CONTAINER_RESP, ret)

    def test_container_show(self):
        headers = {
            'X-Container-Meta-Owner': FAKE_ACCOUNT,
            'x-container-object-count': '1',
            'x-container-bytes-used': '577',
            'x-storage-policy': 'o1--sr-r3'
        }
        resp = {
            'account': FAKE_ACCOUNT,
            'container': 'qaz',
            'object_count': '1',
            'bytes_used': '577',
            'storage_policy': 'o1--sr-r3',
            'properties': {'Owner': FAKE_ACCOUNT},
        }
        self.requests_mock.register_uri(
            'HEAD',
            FAKE_URL + '/qaz',
            headers=headers,
            status_code=204,
        )
        ret = self.api.container_show(container='qaz')
        self.assertEqual(resp, ret)


class TestObject(TestObjectAPIv1):

    def setUp(self):
        super(TestObject, self).setUp()

    @mock.patch('openstackclient.api.object_store_v1.io.open')
    def base_object_create(self, file_contents, mock_open):
        mock_open.read.return_value = file_contents

        headers = {
            'etag': 'youreit',
            'x-trans-id': '1qaz2wsx',
        }
        # TODO(dtroyer): When requests_mock gains the ability to
        #                match against request.body add this check
        #                https://review.opendev.org/127316
        self.requests_mock.register_uri(
            'PUT',
            FAKE_URL + '/qaz/counter.txt',
            headers=headers,
            # body=file_contents,
            status_code=201,
        )
        ret = self.api.object_create(
            container='qaz',
            object='counter.txt',
        )
        data = {
            'account': FAKE_ACCOUNT,
            'container': 'qaz',
            'object': 'counter.txt',
            'etag': 'youreit',
            'x-trans-id': '1qaz2wsx',
        }
        self.assertEqual(data, ret)

    def test_object_create(self):
        self.base_object_create('111\n222\n333\n')
        self.base_object_create(bytes([0x31, 0x00, 0x0d, 0x0a, 0x7f, 0xff]))

    def test_object_delete(self):
        self.requests_mock.register_uri(
            'DELETE',
            FAKE_URL + '/qaz/wsx',
            status_code=204,
        )
        ret = self.api.object_delete(
            container='qaz',
            object='wsx',
        )
        self.assertIsNone(ret)

    def test_object_list_no_options(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '/qaz',
            json=LIST_OBJECT_RESP,
            status_code=200,
        )
        ret = self.api.object_list(container='qaz')
        self.assertEqual(LIST_OBJECT_RESP, ret)

    def test_object_list_delimiter(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '/qaz?delimiter=%7C',
            json=LIST_OBJECT_RESP,
            status_code=200,
        )
        ret = self.api.object_list(
            container='qaz',
            delimiter='|',
        )
        self.assertEqual(LIST_OBJECT_RESP, ret)

    def test_object_list_prefix(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '/qaz?prefix=foo%2f',
            json=LIST_OBJECT_RESP,
            status_code=200,
        )
        ret = self.api.object_list(
            container='qaz',
            prefix='foo/',
        )
        self.assertEqual(LIST_OBJECT_RESP, ret)

    def test_object_list_marker_limit_end(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '/qaz?marker=next&limit=2&end_marker=stop',
            json=LIST_CONTAINER_RESP,
            status_code=200,
        )
        ret = self.api.object_list(
            container='qaz',
            marker='next',
            limit=2,
            end_marker='stop',
        )
        self.assertEqual(LIST_CONTAINER_RESP, ret)

#     def test_list_objects_full_listing(self):
#         sess = self.app.client_manager.session
#
#         def side_effect(*args, **kwargs):
#             rv = sess.get().json.return_value
#             sess.get().json.return_value = []
#             sess.get().json.side_effect = None
#             return rv
#
#         resp = [{'name': 'is-name'}]
#         sess.get().json.return_value = resp
#         sess.get().json.side_effect = side_effect
#
#         data = lib_object.list_objects(
#             sess,
#             fake_url,
#             fake_container,
#             full_listing=True,
#         )
#
#         # Check expected values
#         sess.get.assert_called_with(
#             fake_url + '/' + fake_container,
#             params={
#                 'format': 'json',
#                 'marker': 'is-name',
#             }
#         )
#         self.assertEqual(resp, data)

    def test_object_show(self):
        headers = {
            'content-type': 'text/alpha',
            'content-length': '577',
            'last-modified': '20130101',
            'etag': 'qaz',
            'x-container-meta-owner': FAKE_ACCOUNT,
            'x-object-meta-wife': 'Wilma',
            'x-object-meta-Husband': 'fred',
            'x-tra-header': 'yabba-dabba-do',
        }
        resp = {
            'account': FAKE_ACCOUNT,
            'container': 'qaz',
            'object': FAKE_OBJECT,
            'content-type': 'text/alpha',
            'content-length': '577',
            'last-modified': '20130101',
            'etag': 'qaz',
            'properties': {'wife': 'Wilma',
                           'Husband': 'fred'},
        }
        self.requests_mock.register_uri(
            'HEAD',
            FAKE_URL + '/qaz/' + FAKE_OBJECT,
            headers=headers,
            status_code=204,
        )
        ret = self.api.object_show(
            container='qaz',
            object=FAKE_OBJECT,
        )
        self.assertEqual(resp, ret)
